# -*- coding: utf-8 -*-
# File: conftest.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for fixtures
"""

from typing import Dict
from pathlib import Path

import numpy as np

from pytest import fixture

from deepdoctection.utils.systools import get_package_path
from deepdoctection.utils.settings import LayoutType, ObjectTypes
from deepdoctection.utils.detection_types import ImageType


@fixture(name="path_to_d2_frcnn_yaml")
def fixture_path_to_d2_frcnn_yaml() -> Path:
    """
    path to d2 frcnn yaml file
    """
    return get_package_path() / "configs/d2/layout/CASCADE_RCNN_R_50_FPN_GN.yaml"


@fixture(name="categories")
def fixture_categories() -> Dict[str, ObjectTypes]:
    """
    Categories as Dict
    """
    return {"1": LayoutType.text, "2": LayoutType.title, "3": LayoutType.table,
            "4": LayoutType.figure, "5": LayoutType.list}


@fixture(name="np_image")
def fixture_np_image() -> ImageType:
    """
    np_array image
    """
    return np.ones([4, 6, 3], dtype=np.float32)



# -*- coding: utf-8 -*-
# File: stats.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Dataflows for calculating statistical values of the underlying dataset
"""
from typing import Any, Optional, Tuple, Union

import numpy as np
import numpy.typing as npt

from ..utils.logger import logger
from ..utils.tqdm import get_tqdm
from .base import DataFlow, ProxyDataFlow


class MeanFromDataFlow(ProxyDataFlow):
    """
    Get the mean of some dataflow. Takes a component from a dataflow and calculates iteratively the mean.

    **Example:**

            df: some dataflow
            MeanFromDataFlow(df).start() if you want to put MeanFromDataFlow at the end of a dataflow
            df: some dataflow
            df = MeanFromDataFlow(df)

        is also possible. Testing with the progress bar will stop once the requested size has been reached.
    """

    def __init__(
        self,
        df: DataFlow,
        axis: Optional[Union[int, Tuple[int], Tuple[int, int], Tuple[int, int, int]]] = None,
        key: Optional[str] = None,
        max_datapoints: Optional[int] = None,
    ):
        """
        :param df: the dataflow to test
        :param axis: The axis along which to calculate the mean. It will always calculate the mean along
                     the dataflow length axis, which is the 0th axis. E.g. for calculating the mean of an image dataset
                     use

                         MeanFromDataFlow(df,key="image",axis=(0,1,2)).start()

        :param key: The datapoint key giving the values to evaluate the mean. If None it tries to use
                    the whole datapoint.
        :param max_datapoints: Will stop considering datapoints with index>max_datapoints.
        """

        super().__init__(df)
        if isinstance(axis, int):
            axis = (axis,)
        if axis is not None:
            if 0 not in axis:
                raise ValueError("0-th axis required for calculating mean")
        self.axis = axis
        self.key = key
        self.max_datapoints = max_datapoints
        self.mean: npt.NDArray[Any]
        self._reset_called = False

    def reset_state(self) -> None:
        self._reset_called = True
        super().reset_state()

    def __iter__(self) -> Any:
        """
        Will run the mean calculation at the until max_datapoints (if given) has been reached, then it will
        produce data normally.
        """
        self.start()
        yield from self.df

    def start(self) -> npt.NDArray[Any]:
        """
        Start calculating the mean with a progress bar.
        """

        if not self._reset_called:
            self.df.reset_state()
        itr = self.df.__iter__()

        logger.info("Calculating mean")

        len_df: Optional[int]
        try:
            len_df = len(self.df)
        except NotImplementedError:
            len_df = None
        if len_df is not None and self.max_datapoints is not None:
            len_df = min(len_df, self.max_datapoints)

        with get_tqdm(total=len_df) as status_bar:
            n = None
            for n, dp in enumerate(itr, 1):

                if isinstance(dp, dict):
                    assert isinstance(self.key, str), self.key
                    val = dp[self.key]
                elif isinstance(dp, list):
                    val = dp
                else:
                    assert isinstance(self.key, str), self.key
                    val = getattr(dp, self.key)

                if isinstance(val, (tuple, list)):
                    val = np.asarray(val)

                if n == 1:
                    val_0_ndim = val.ndim
                    inner_axis = None
                    if self.axis is None:
                        self.mean = np.zeros(())
                    else:
                        inner_axis = tuple(np.subtract(self.axis, tuple((1 for _ in self.axis))))[1:]
                        mean_shape = tuple(val.shape[i] for i in range(len(val.shape)) if i not in inner_axis)
                        self.mean = np.zeros(mean_shape)

                if val.ndim == val_0_ndim:
                    x = np.mean(val, axis=self.axis if inner_axis is None else inner_axis)
                    self.mean += (x - self.mean) / n

                status_bar.update()  # type: ignore
                if self.max_datapoints is not None:
                    if n == self.max_datapoints:
                        break

        logger.info("Mean from %s datapoints along axis %s: %s", n, self.axis, self.mean)

        return self.mean


class StdFromDataFlow(ProxyDataFlow):
    """
    Gets the standard deviation of some dataflow. Takes a component from a dataflow and calculates iteratively
    the standard deviation.

    **Example:**

            df= ...
            StdFromDataFlow(df).start()

        if you want to put  StdFromDataFlow at the end of a dataflow

            df: some dataflow
            df = StdFromDataFlow(df)

        is also possible. The testing with the progress bar will stop once the requested size has been reached.
    """

    def __init__(
        self,
        df: DataFlow,
        axis: Optional[Union[int, Tuple[int], Tuple[int, int], Tuple[int, int, int]]] = None,
        key: Optional[str] = None,
        max_datapoints: Optional[int] = None,
    ):
        """
        :param df: the dataflow to test
        :param axis: The axis along which to calculate the mean. It will always calculate the std along
                     the dataflow length axis, which is the 0th axis. E.g. for calculating the mean of an image dataset
                     use

                         StdFromDataFlow(df,key="image",axis=(0,1,2)).start()

        :param key: The datapoint key giving the values to evaluate the std. If None it tries to use
                    the whole datapoint.
        :param max_datapoints: Will stop considering datapoints with index>max_datapoints.
        """

        super().__init__(df)
        if isinstance(axis, int):
            axis = (axis,)
        if axis is not None:
            if 0 not in axis:
                raise ValueError("0-th axis required for calculating std")
        self.axis = axis
        self.key = key
        self.max_datapoints = max_datapoints
        self.std: npt.NDArray[Any]
        self._reset_called = False

    def reset_state(self) -> None:
        self._reset_called = True
        super().reset_state()

    def __iter__(self) -> Any:
        """
        Will run the mean calculation at the until max_datapoints (if given) has been reached, then it will
        produce data normally.
        """
        self.start()
        yield from self.df

    def start(self) -> npt.NDArray[Any]:
        """
        Start calculating the mean with a progress bar.
        """
        len_df: Optional[int]

        if not self._reset_called:
            self.df.reset_state()
        itr = self.df.__iter__()

        logger.info("Calculating standard deviation")
        try:
            len_df = len(self.df)
        except NotImplementedError:
            len_df = None
        if len_df is not None and self.max_datapoints is not None:
            len_df = min(len_df, self.max_datapoints)
        n = None
        with get_tqdm(total=len_df) as status_bar:
            for n, dp in enumerate(itr, 1):

                if isinstance(dp, dict):
                    assert isinstance(self.key, str), self.key
                    val = dp[self.key]
                elif isinstance(dp, list):
                    val = dp
                else:
                    assert isinstance(self.key, str), self.key
                    val = getattr(dp, self.key)

                if isinstance(val, (tuple, list)):
                    val = np.asarray(val)

                if n == 1:
                    val_0_ndim = val.ndim
                    inner_axis = None
                    if self.axis is None:
                        self.std = np.zeros(())
                        ex, ex2 = np.zeros(()), np.zeros(())
                    else:
                        inner_axis = tuple(np.subtract(self.axis, tuple((1 for _ in self.axis))))[1:]
                        std_shape = tuple(val.shape[i] for i in range(len(val.shape)) if i not in inner_axis)
                        self.std = np.zeros(std_shape)
                        ex, ex2 = np.zeros(std_shape), np.zeros(std_shape)

                if val.ndim == val_0_ndim:
                    x = np.mean(val, axis=self.axis if inner_axis is None else inner_axis)
                    if n == 1:
                        k = x

                    ex += x - k
                    ex2 += (x - k) * (x - k)

                status_bar.update()  # type: ignore
                if self.max_datapoints is not None:
                    if n == self.max_datapoints:
                        break

            var = (ex2 - (ex * ex) / n) / (n - 1)
            self.std = np.sqrt(var)

        logger.info("Standard deviation from %s datapoints along axis %s: %s", n, self.axis, self.std)

        return self.std



# -*- coding: utf-8 -*-
# File: fastlang.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Deepdoctection wrappers for fasttext language detection models
"""
from copy import copy
from typing import List, Mapping

from ..utils.file_utils import Requirement, fasttext_available, get_fasttext_requirement
from ..utils.settings import TypeOrStr
from .base import DetectionResult, LanguageDetector, PredictorBase

if fasttext_available():
    from fasttext import load_model  # type: ignore


class FasttextLangDetector(LanguageDetector):
    """
    Fasttext language detector wrapper. Two models provided in the fasttext library can be used to identify languages.
    The background to the models can be found in the works:

    [1] Joulin A, Grave E, Bojanowski P, Mikolov T, Bag of Tricks for Efficient Text Classification

    [2] Joulin A, Grave E, Bojanowski P, Douze M, Jégou H, Mikolov T, FastText.zip: Compressing text classification
        models

    The models are distributed under the Creative Commons Attribution-Share-Alike License 3.0.
    (<https://creativecommons.org/licenses/by-sa/3.0/>)

    When loading the models via the ModelCatalog, the original and unmodified models are used.

        path_weights = ModelCatalog.get_full_path_weights("fasttext/lid.176.bin")
        profile = ModelCatalog.get_profile("fasttext/lid.176.bin")
        lang_detector = FasttextLangDetector(path_weights,profile.categories)
        detection_result = lang_detector.predict("some text in some language")

    """

    def __init__(self, path_weights: str, categories: Mapping[str, TypeOrStr]):
        """
        :param path_weights: path to model weights
        :param categories: A dict with the model output label and value. We use as convention the ISO 639-2 language
                           code.
        """

        self.name = "fasttest_lang_detector"
        self.path_weights = path_weights
        self.model = load_model(self.path_weights)
        self.categories = copy(categories)  # type: ignore

    def predict(self, text_string: str) -> DetectionResult:
        output = self.model.predict(text_string)
        return DetectionResult(text=self.categories[output[0][0]], score=output[1][0])

    @classmethod
    def get_requirements(cls) -> List[Requirement]:
        return [get_fasttext_requirement()]

    def clone(self) -> PredictorBase:
        return self.__class__(self.path_weights, self.categories)



# -*- coding: utf-8 -*-
# File: test_anngen.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module pipe.anngen
"""

from typing import List

from pytest import mark

from deepdoctection.datapoint import Image, ImageAnnotation
from deepdoctection.extern.base import DetectionResult
from deepdoctection.pipe.anngen import DatapointManager
from deepdoctection.utils.settings import get_type


class TestDatapointManager:
    """
    Testing DatapointManager functions
    """

    @staticmethod
    @mark.basic
    def test_set_image_annotation(
        dp_image: Image, layout_detect_results: List[DetectionResult], layout_annotations: List[ImageAnnotation]
    ) -> None:
        """
        test set_image_annotation
        """
        # Arrange
        dp_manager = DatapointManager()
        dp_manager.datapoint = dp_image

        # Act
        ann_id = dp_manager.set_image_annotation(layout_detect_results[0])

        # Assert
        assert ann_id is not None
        ann = dp_manager.datapoint.get_annotation()
        assert ann[0] == layout_annotations[0]
        assert ann_id == layout_annotations[0].annotation_id
        assert dp_manager._cache_anns[ann_id] == layout_annotations[0]  # pylint: disable=W0212

    @staticmethod
    @mark.basic
    def test_set_image_annotation_with_image(dp_image: Image, layout_detect_results: List[DetectionResult]) -> None:
        """
        test set_image_annotation with image_ann_to_image
        """
        # Arrange
        dp_manager = DatapointManager()
        dp_manager.datapoint = dp_image

        # Act
        dp_manager.set_image_annotation(layout_detect_results[0], to_image=True)

        # Assert
        ann = dp_manager.datapoint.get_annotation()
        assert ann[0].image is not None
        assert ann[0].bounding_box == ann[0].image.get_embedding(dp_image.image_id)

    @staticmethod
    @mark.basic
    def test_set_image_annotation_to_image_ann(dp_image: Image, layout_detect_results: List[DetectionResult]) -> None:
        """
        test set_image_annotation with ann to image ann
        """

        # Arrange
        dp_manager = DatapointManager()
        dp_manager.datapoint = dp_image
        img_ann_id = dp_manager.set_image_annotation(layout_detect_results[0], to_image=True)

        # Act
        ann_id = dp_manager.set_image_annotation(layout_detect_results[2], to_annotation_id=img_ann_id)

        # Assert
        ann = dp_manager.datapoint.get_annotation(annotation_ids=ann_id)
        img_ann = dp_manager.datapoint.get_annotation(annotation_ids=img_ann_id)[0]

        assert img_ann.image is not None
        ann_from_img_ann = img_ann.image.get_annotation(annotation_ids=ann_id)
        assert ann == ann_from_img_ann
        assert dp_manager._cache_anns[ann_id]  # type:ignore # pylint: disable=W0212

    @staticmethod
    @mark.basic
    def test_set_category_annotation(dp_image: Image, layout_detect_results: List[DetectionResult]) -> None:
        """
        test set_category_annotation
        """

        # Arrange
        dp_manager = DatapointManager()
        dp_manager.datapoint = dp_image
        ann_id = dp_manager.set_image_annotation(layout_detect_results[0])

        # Act
        assert ann_id is not None
        dp_manager.set_category_annotation(get_type("foo"), 5, get_type("FOO"), ann_id, 0.8)

        # Assert
        ann = dp_manager.datapoint.get_annotation(annotation_ids=ann_id)
        cat_ann = ann[0].get_sub_category(get_type("FOO"))

        assert cat_ann.category_id == "5"
        assert cat_ann.score == 0.8
        assert cat_ann.category_name == "foo"

    @staticmethod
    @mark.basic
    def test_set_container_annotation(dp_image: Image, layout_detect_results: List[DetectionResult]) -> None:
        """
        test set_container_annotation
        """

        # Arrange
        dp_manager = DatapointManager()
        dp_manager.datapoint = dp_image
        ann_id = dp_manager.set_image_annotation(layout_detect_results[0])

        # Act
        assert ann_id is not None
        cont_ann_id = dp_manager.set_container_annotation(
            get_type("foo"), 5, get_type("FOO"), ann_id, "hello world", 0.8
        )

        # Assert
        ann = dp_manager.datapoint.get_annotation(annotation_ids=ann_id)
        cont_ann = ann[0].get_sub_category(get_type("FOO"))

        assert cont_ann.category_id == "5"
        assert cont_ann.score == 0.8
        assert cont_ann.category_name == "foo"
        assert cont_ann.value == "hello world"  # type: ignore
        assert cont_ann.annotation_id == cont_ann_id

    @staticmethod
    @mark.basic
    def test_summary_annotation(dp_image: Image, layout_detect_results: List[DetectionResult]) -> None:
        """
        test summary_annotation
        """

        # Arrange
        dp_manager = DatapointManager()
        dp_manager.datapoint = dp_image
        ann_id = dp_manager.set_image_annotation(layout_detect_results[0], to_image=True)

        # Act
        summ_id_1 = dp_manager.set_summary_annotation(get_type("foo"), get_type("foo"), 1)
        summ_id_2 = dp_manager.set_summary_annotation(get_type("bak"), get_type("bak"), 2, annotation_id=ann_id)
        ann = dp_manager.datapoint.get_annotation(annotation_ids=ann_id)

        # Assert
        cat_1 = dp_manager.datapoint.summary.get_sub_category(get_type("foo"))  # type: ignore
        assert cat_1.annotation_id == summ_id_1
        assert cat_1.category_name == "foo"
        assert cat_1.category_id == "1"

        cat_2 = ann[0].image.summary.get_sub_category(get_type("bak"))  # type: ignore
        assert cat_2.annotation_id == summ_id_2
        assert cat_2.category_name == "bak"
        assert cat_2.category_id == "2"



# -*- coding: utf-8 -*-
# File: model.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for ModelCatalog and ModelDownloadManager
"""

import os
from copy import copy
from dataclasses import asdict, dataclass, field
from typing import Any, Dict, List, Mapping, Optional, Union

from huggingface_hub import cached_download, hf_hub_url  # type: ignore
from tabulate import tabulate
from termcolor import colored

from ..utils.fs import download
from ..utils.logger import log_once, logger
from ..utils.settings import CellType, Languages, LayoutType, ObjectTypes
from ..utils.systools import get_configs_dir_path, get_weights_dir_path

__all__ = ["ModelCatalog", "ModelDownloadManager", "print_model_infos", "ModelProfile"]


@dataclass
class ModelProfile:
    """
    Class for model profile. Add for each model one ModelProfile to the ModelCatalog
    """

    name: str
    description: str

    size: List[int]
    tp_model: bool = field(default=False)
    config: Optional[str] = field(default=None)
    preprocessor_config: Optional[str] = field(default=None)
    hf_repo_id: Optional[str] = field(default=None)
    hf_model_name: Optional[str] = field(default=None)
    hf_config_file: Optional[List[str]] = field(default=None)
    urls: Optional[List[str]] = field(default=None)
    categories: Optional[Dict[str, ObjectTypes]] = field(default=None)
    model_wrapper: Optional[str] = field(default=None)

    def as_dict(self) -> Dict[str, Any]:
        """
        returns a dict of the dataclass
        """
        return asdict(self)


class ModelCatalog:
    """
    Catalog of some pre-trained models. The associated config file is available as well.

    To get an overview of all registered models

        print(ModelCatalog.get_model_list())

    To get a model card for some specific model:

        profile = ModelCatalog.get_profile("layout/model-800000_inf_only.data-00000-of-00001")
        print(profile.description)

    Some models will have their weights and configs stored in the cache. To instantiate predictors one will sometimes
    need their path. Use

        path_weights = ModelCatalog.get_full_path_configs("layout/model-800000_inf_only.data-00000-of-00001")
        path_configs = ModelCatalog.get_full_path_weights("layout/model-800000_inf_only.data-00000-of-00001")

    To register a new model

        ModelCatalog.get_full_path_configs("my_new_model")
    """

    CATALOG: Dict[str, ModelProfile] = {
        "layout/model-800000_inf_only.data-00000-of-00001": ModelProfile(
            name="layout/model-800000_inf_only.data-00000-of-00001",
            description="Tensorpack layout model for inference purposes trained on Publaynet",
            config="dd/tp/conf_frcnn_layout.yaml",
            size=[274552244, 7907],
            tp_model=True,
            hf_repo_id="deepdoctection/tp_casc_rcnn_X_32xd4_50_FPN_GN_2FC_publaynet_inference_only",
            hf_model_name="model-800000_inf_only",
            hf_config_file=["conf_frcnn_layout.yaml"],
            categories={
                "1": LayoutType.text,
                "2": LayoutType.title,
                "3": LayoutType.list,
                "4": LayoutType.table,
                "5": LayoutType.figure,
            },
            model_wrapper="TPFrcnnDetector",
        ),
        "cell/model-1800000_inf_only.data-00000-of-00001": ModelProfile(
            name="cell/model-1800000_inf_only.data-00000-of-00001",
            description="Tensorpack cell detection model for inference purposes trained on Pubtabnet",
            config="dd/tp/conf_frcnn_cell.yaml",
            size=[274503056, 8056],
            tp_model=True,
            hf_repo_id="deepdoctection/tp_casc_rcnn_X_32xd4_50_FPN_GN_2FC_pubtabnet_c_inference_only",
            hf_model_name="model-1800000_inf_only",
            hf_config_file=["conf_frcnn_cell.yaml"],
            categories={"1": LayoutType.cell},
            model_wrapper="TPFrcnnDetector",
        ),
        "item/model-1620000_inf_only.data-00000-of-00001": ModelProfile(
            name="item/model-1620000_inf_only.data-00000-of-00001",
            description="Tensorpack row/column detection model for inference purposes trained on Pubtabnet",
            config="dd/tp/conf_frcnn_rows.yaml",
            size=[274515344, 7904],
            tp_model=True,
            hf_repo_id="deepdoctection/tp_casc_rcnn_X_32xd4_50_FPN_GN_2FC_pubtabnet_rc_inference_only",
            hf_model_name="model-1620000_inf_only",
            hf_config_file=["conf_frcnn_rows.yaml"],
            categories={"1": LayoutType.row, "2": LayoutType.column},
            model_wrapper="TPFrcnnDetector",
        ),
        "item/model-1620000.data-00000-of-00001": ModelProfile(
            name="item/model-1620000.data-00000-of-00001",
            description="Tensorpack row/column detection model trained on Pubtabnet",
            config="dd/tp/conf_frcnn_rows.yaml",
            size=[823546048, 25787],
            tp_model=True,
            hf_repo_id="deepdoctection/tp_casc_rcnn_X_32xd4_50_FPN_GN_2FC_pubtabnet_rc",
            hf_model_name="model-1620000",
            hf_config_file=["conf_frcnn_rows.yaml"],
            categories={"1": LayoutType.row, "2": LayoutType.column},
            model_wrapper="TPFrcnnDetector",
        ),
        "layout/model-800000.data-00000-of-00001": ModelProfile(
            name="layout/model-800000.data-00000-of-00001",
            description="Tensorpack layout detection model trained on Publaynet",
            config="dd/tp/conf_frcnn_layout.yaml",
            size=[823656748, 25796],
            tp_model=True,
            hf_repo_id="deepdoctection/tp_casc_rcnn_X_32xd4_50_FPN_GN_2FC_publaynet",
            hf_model_name="model-800000",
            hf_config_file=["conf_frcnn_layout.yaml"],
            categories={
                "1": LayoutType.text,
                "2": LayoutType.title,
                "3": LayoutType.list,
                "4": LayoutType.table,
                "5": LayoutType.figure,
            },
            model_wrapper="TPFrcnnDetector",
        ),
        "cell/model-1800000.data-00000-of-00001": ModelProfile(
            name="cell/model-1800000.data-00000-of-00001",
            description="Tensorpack cell detection model trained on Pubtabnet",
            config="dd/tp/conf_frcnn_cell.yaml",
            size=[823509160, 25905],
            tp_model=True,
            hf_repo_id="deepdoctection/tp_casc_rcnn_X_32xd4_50_FPN_GN_2FC_pubtabnet_c",
            hf_model_name="model-1800000",
            hf_config_file=["conf_frcnn_cell.yaml"],
            categories={"1": LayoutType.cell},
            model_wrapper="TPFrcnnDetector",
        ),
        "layout/d2_model-800000-layout.pkl": ModelProfile(
            name="layout/d2_model-800000-layout.pkl",
            description="Detectron2 layout detection model trained on Publaynet",
            config="dd/d2/layout/CASCADE_RCNN_R_50_FPN_GN.yaml",
            size=[274568239],
            tp_model=False,
            hf_repo_id="deepdoctection/d2_casc_rcnn_X_32xd4_50_FPN_GN_2FC_publaynet_inference_only",
            hf_model_name="d2_model-800000-layout.pkl",
            hf_config_file=["Base-RCNN-FPN.yaml", "CASCADE_RCNN_R_50_FPN_GN.yaml"],
            categories={
                "1": LayoutType.text,
                "2": LayoutType.title,
                "3": LayoutType.list,
                "4": LayoutType.table,
                "5": LayoutType.figure,
            },
            model_wrapper="D2FrcnnDetector",
        ),
        "layout/d2_model_0829999_layout_inf_only.pt": ModelProfile(
            name="layout/d2_model_0829999_layout_inf_only.pt",
            description="Detectron2 layout detection model trained on Publaynet",
            config="dd/d2/layout/CASCADE_RCNN_R_50_FPN_GN.yaml",
            size=[274632215],
            tp_model=False,
            hf_repo_id="deepdoctection/d2_casc_rcnn_X_32xd4_50_FPN_GN_2FC_publaynet_inference_only",
            hf_model_name="d2_model_0829999_layout_inf_only.pt",
            hf_config_file=["Base-RCNN-FPN.yaml", "CASCADE_RCNN_R_50_FPN_GN.yaml"],
            categories={
                "1": LayoutType.text,
                "2": LayoutType.title,
                "3": LayoutType.list,
                "4": LayoutType.table,
                "5": LayoutType.figure,
            },
            model_wrapper="D2FrcnnDetector",
        ),
        "layout/d2_model_0829999_layout.pth": ModelProfile(
            name="layout/d2_model_0829999_layout.pth",
            description="Detectron2 layout detection model trained on Publaynet. Checkpoint for resuming training",
            config="dd/d2/layout/CASCADE_RCNN_R_50_FPN_GN.yaml",
            size=[548377327],
            tp_model=False,
            hf_repo_id="deepdoctection/d2_casc_rcnn_X_32xd4_50_FPN_GN_2FC_publaynet_inference_only",
            hf_model_name="d2_model_0829999_layout.pth",
            hf_config_file=["Base-RCNN-FPN.yaml", "CASCADE_RCNN_R_50_FPN_GN.yaml"],
            categories={
                "1": LayoutType.text,
                "2": LayoutType.title,
                "3": LayoutType.list,
                "4": LayoutType.table,
                "5": LayoutType.figure,
            },
            model_wrapper="D2FrcnnDetector",
        ),
        "cell/d2_model-1800000-cell.pkl": ModelProfile(
            name="cell/d2_model-1800000-cell.pkl",
            description="Detectron2 cell detection inference only model trained on Pubtabnet",
            config="dd/d2/cell/CASCADE_RCNN_R_50_FPN_GN.yaml",
            size=[274519039],
            tp_model=False,
            hf_repo_id="deepdoctection/d2_casc_rcnn_X_32xd4_50_FPN_GN_2FC_pubtabnet_c_inference_only",
            hf_model_name="d2_model-1800000-cell.pkl",
            hf_config_file=["Base-RCNN-FPN.yaml", "CASCADE_RCNN_R_50_FPN_GN.yaml"],
            categories={"1": LayoutType.cell},
            model_wrapper="D2FrcnnDetector",
        ),
        "cell/d2_model_1849999_cell_inf_only.pt": ModelProfile(
            name="cell/d2_model_1849999_cell_inf_only.pt",
            description="Detectron2 cell detection inference only model trained on Pubtabnet",
            config="dd/d2/cell/CASCADE_RCNN_R_50_FPN_GN.yaml",
            size=[274583063],
            tp_model=False,
            hf_repo_id="deepdoctection/d2_casc_rcnn_X_32xd4_50_FPN_GN_2FC_pubtabnet_c_inference_only",
            hf_model_name="d2_model_1849999_cell_inf_only.pt",
            hf_config_file=["Base-RCNN-FPN.yaml", "CASCADE_RCNN_R_50_FPN_GN.yaml"],
            categories={"1": LayoutType.cell},
            model_wrapper="D2FrcnnDetector",
        ),
        "cell/d2_model_1849999_cell.pth": ModelProfile(
            name="cell/d2_model_1849999_cell.pth",
            description="Detectron2 cell detection inference only model trained on Pubtabnet",
            config="dd/d2/cell/CASCADE_RCNN_R_50_FPN_GN.yaml",
            size=[548279023],
            tp_model=False,
            hf_repo_id="deepdoctection/d2_casc_rcnn_X_32xd4_50_FPN_GN_2FC_pubtabnet_c_inference_only",
            hf_model_name="cell/d2_model_1849999_cell.pth",
            hf_config_file=["Base-RCNN-FPN.yaml", "CASCADE_RCNN_R_50_FPN_GN.yaml"],
            categories={"1": LayoutType.cell},
            model_wrapper="D2FrcnnDetector",
        ),
        "item/d2_model-1620000-item.pkl": ModelProfile(
            name="item/d2_model-1620000-item.pkl",
            description="Detectron2 item detection inference only model trained on Pubtabnet",
            config="dd/d2/item/CASCADE_RCNN_R_50_FPN_GN.yaml",
            size=[274531339],
            tp_model=False,
            hf_repo_id="deepdoctection/d2_casc_rcnn_X_32xd4_50_FPN_GN_2FC_pubtabnet_rc_inference_only",
            hf_model_name="d2_model-1620000-item.pkl",
            hf_config_file=["Base-RCNN-FPN.yaml", "CASCADE_RCNN_R_50_FPN_GN.yaml"],
            categories={"1": LayoutType.row, "2": LayoutType.column},
            model_wrapper="D2FrcnnDetector",
        ),
        "item/d2_model_1639999_item.pth": ModelProfile(
            name="item/d2_model_1639999_item.pth",
            description="Detectron2 item detection model trained on Pubtabnet",
            config="dd/d2/item/CASCADE_RCNN_R_50_FPN_GN.yaml",
            size=[548303599],
            tp_model=False,
            hf_repo_id="deepdoctection/d2_casc_rcnn_X_32xd4_50_FPN_GN_2FC_pubtabnet_rc_inference_only",
            hf_model_name="d2_model_1639999_item.pth",
            hf_config_file=["Base-RCNN-FPN.yaml", "CASCADE_RCNN_R_50_FPN_GN.yaml"],
            categories={"1": LayoutType.row, "2": LayoutType.column},
            model_wrapper="D2FrcnnDetector",
        ),
        "item/d2_model_1639999_item_inf_only.pt": ModelProfile(
            name="item/d2_model_1639999_item_inf_only.pt",
            description="Detectron2 item detection model inference only trained on Pubtabnet",
            config="dd/d2/item/CASCADE_RCNN_R_50_FPN_GN.yaml",
            size=[274595351],
            tp_model=False,
            hf_repo_id="deepdoctection/d2_casc_rcnn_X_32xd4_50_FPN_GN_2FC_pubtabnet_rc_inference_only",
            hf_model_name="d2_model_1639999_item_inf_only.pt",
            hf_config_file=["Base-RCNN-FPN.yaml", "CASCADE_RCNN_R_50_FPN_GN.yaml"],
            categories={"1": LayoutType.row, "2": LayoutType.column},
            model_wrapper="D2FrcnnDetector",
        ),
        "microsoft/layoutlm-base-uncased/pytorch_model.bin": ModelProfile(
            name="microsoft/layoutlm-base-uncased/pytorch_model.bin",
            description="LayoutLM is a simple but effective pre-training method of text and layout for document image"
            " understanding and information extraction tasks, such as form understanding and receipt"
            " understanding. LayoutLM archived the SOTA results on multiple datasets. This model does not"
            "contain any head and has to be fine tuned on a downstream task. This is model has been trained "
            "on 11M documents for 2 epochs.  Configuration: 12-layer, 768-hidden, 12-heads, 113M parameters",
            size=[453093832],
            tp_model=False,
            config="microsoft/layoutlm-base-uncased/config.json",
            hf_repo_id="microsoft/layoutlm-base-uncased",
            hf_model_name="pytorch_model.bin",
            hf_config_file=["config.json"],
        ),
        "microsoft/layoutlm-large-uncased/pytorch_model.bin": ModelProfile(
            name="microsoft/layoutlm-large-uncased/pytorch_model.bin",
            description="LayoutLM is a simple but effective pre-training method of text and layout for document image"
            " understanding and information extraction tasks, such as form understanding and receipt"
            " understanding. LayoutLM archived the SOTA results on multiple datasets. This model does not"
            "contain any head and has to be fine tuned on a downstream task. This is model has been trained"
            " on 11M documents for 2 epochs.  Configuration: 24-layer, 1024-hidden, 16-heads, 343M parameters",
            size=[1361845448],
            tp_model=False,
            config="microsoft/layoutlm-large-uncased/config.json",
            hf_repo_id="microsoft/layoutlm-large-uncased",
            hf_model_name="pytorch_model.bin",
            hf_config_file=["config.json"],
        ),
        "microsoft/layoutlmv2-base-uncased/pytorch_model.bin": ModelProfile(
            name="microsoft/layoutlmv2-base-uncased/pytorch_model.bin",
            description="LayoutLMv2 is an improved version of LayoutLM with new pre-training tasks to model the"
            " interaction among text, layout, and image in a single multi-modal framework. It outperforms"
            " strong baselines and achieves new state-of-the-art results on a wide variety of downstream"
            " visually-rich document understanding tasks, including , including FUNSD (0.7895 → 0.8420),"
            " CORD (0.9493 → 0.9601), SROIE (0.9524 → 0.9781), Kleister-NDA (0.834 → 0.852), RVL-CDIP"
            " (0.9443 → 0.9564), and DocVQA (0.7295 → 0.8672). The license is cc-by-nc-sa-4.0",
            size=[802243295],
            tp_model=False,
            config="microsoft/layoutlmv2-base-uncased/config.json",
            hf_repo_id="microsoft/layoutlmv2-base-uncased",
            hf_model_name="pytorch_model.bin",
            hf_config_file=["config.json"],
        ),
        "microsoft/layoutxlm-base/pytorch_model.bin": ModelProfile(
            name="microsoft/layoutxlm-base/pytorch_model.bin",
            description="Multimodal pre-training with text, layout, and image has achieved SOTA performance for "
            "visually-rich document understanding tasks recently, which demonstrates the great potential"
            " for joint learning across different modalities. In this paper, we present LayoutXLM, a"
            " multimodal pre-trained model for multilingual document understanding, which aims to bridge"
            " the language barriers for visually-rich document understanding. To accurately evaluate"
            " LayoutXLM, we also introduce a multilingual form understanding benchmark dataset named XFUN,"
            " which includes form understanding samples in 7 languages (Chinese, Japanese, Spanish, French,"
            " Italian, German, Portuguese), and key-value pairs are manually labeled for each language."
            " Experiment results show that the LayoutXLM model has significantly outperformed the existing"
            " SOTA cross-lingual pre-trained models on the XFUN dataset. The license is cc-by-nc-sa-4.0",
            size=[1476537178],
            tp_model=False,
            config="microsoft/layoutxlm-base/config.json",
            hf_repo_id="microsoft/layoutxlm-base",
            hf_model_name="pytorch_model.bin",
            hf_config_file=["config.json"],
        ),
        "microsoft/layoutlmv3-base/pytorch_model.bin": ModelProfile(
            name="microsoft/layoutlmv3-base/pytorch_model.bin",
            description="LayoutLMv3 is a pre-trained multimodal Transformer for Document AI with unified text and"
            " image masking. The simple unified architecture and training objectives make LayoutLMv3 a"
            " general-purpose pre-trained model. For example, LayoutLMv3 can be fine-tuned for both"
            " text-centric tasks, including form understanding, receipt understanding, and document"
            " visual question answering, and image-centric tasks such as document image classification"
            " and document layout analysis. The license is cc-by-nc-sa-4.0",
            size=[501380823],
            tp_model=False,
            config="microsoft/layoutlmv3-base/config.json",
            hf_repo_id="microsoft/layoutlmv3-base",
            hf_model_name="pytorch_model.bin",
            hf_config_file=["config.json"],
        ),
        "microsoft/table-transformer-detection/pytorch_model.bin": ModelProfile(
            name="microsoft/table-transformer-detection/pytorch_model.bin",
            description="Table Transformer (DETR) model trained on PubTables1M. It was introduced in the paper "
            "PubTables-1M: Towards Comprehensive Table Extraction From Unstructured Documents by Smock et "
            "al. This model is devoted to table detection",
            size=[115393245],
            tp_model=False,
            config="microsoft/table-transformer-detection/config.json",
            preprocessor_config="microsoft/table-transformer-detection/preprocessor_config.json",
            hf_repo_id="microsoft/table-transformer-detection",
            hf_model_name="pytorch_model.bin",
            hf_config_file=["config.json", "preprocessor_config.json"],
            categories={"1": LayoutType.table},
            model_wrapper="HFDetrDerivedDetector",
        ),
        "microsoft/table-transformer-structure-recognition/pytorch_model.bin": ModelProfile(
            name="microsoft/table-transformer-structure-recognition/pytorch_model.bin",
            description="Table Transformer (DETR) model trained on PubTables1M. It was introduced in the paper "
            "PubTables-1M: Towards Comprehensive Table Extraction From Unstructured Documents by Smock et "
            "al. This model is devoted to table structure recognition and assumes to receive a cropped"
            "table as input. It will predict rows, column and spanning cells",
            size=[115509981],
            tp_model=False,
            config="microsoft/table-transformer-structure-recognition/config.json",
            preprocessor_config="microsoft/table-transformer-structure-recognition/preprocessor_config.json",
            hf_repo_id="microsoft/table-transformer-structure-recognition",
            hf_model_name="pytorch_model.bin",
            hf_config_file=["config.json", "preprocessor_config.json"],
            categories={
                "1": LayoutType.table,
                "2": LayoutType.column,
                "3": LayoutType.row,
                "4": CellType.column_header,
                "5": CellType.projected_row_header,
                "6": CellType.spanning,
            },
            model_wrapper="HFDetrDerivedDetector",
        ),
        "fasttext/lid.176.bin": ModelProfile(
            name="fasttext/lid.176.bin",
            description="Fasttext language detection model",
            size=[131266198],
            urls=["https://dl.fbaipublicfiles.com/fasttext/supervised-models/lid.176.bin"],
            categories={
                "__label__en": Languages.english,
                "__label__ru": Languages.russian,
                "__label__de": Languages.german,
                "__label__fr": Languages.french,
                "__label__it": Languages.italian,
                "__label__ja": Languages.japanese,
                "__label__es": Languages.spanish,
                "__label__ceb": Languages.cebuano,
                "__label__tr": Languages.turkish,
                "__label__pt": Languages.portuguese,
                "__label__uk": Languages.ukrainian,
                "__label__eo": Languages.esperanto,
                "__label__pl": Languages.polish,
                "__label__sv": Languages.swedish,
                "__label__nl": Languages.dutch,
                "__label__he": Languages.hebrew,
                "__label__zh": Languages.chinese,
                "__label__hu": Languages.hungarian,
                "__label__ar": Languages.arabic,
                "__label__ca": Languages.catalan,
                "__label__fi": Languages.finnish,
                "__label__cs": Languages.czech,
                "__label__fa": Languages.persian,
                "__label__sr": Languages.serbian,
                "__label__el": Languages.greek,
                "__label__vi": Languages.vietnamese,
                "__label__bg": Languages.bulgarian,
                "__label__ko": Languages.korean,
                "__label__no": Languages.norwegian,
                "__label__mk": Languages.macedonian,
                "__label__ro": Languages.romanian,
                "__label__id": Languages.indonesian,
                "__label__th": Languages.thai,
                "__label__hy": Languages.armenian,
                "__label__da": Languages.danish,
                "__label__ta": Languages.tamil,
                "__label__hi": Languages.hindi,
                "__label__hr": Languages.croatian,
                "__label__sh": Languages.not_defined,
                "__label__be": Languages.belarusian,
                "__label__ka": Languages.georgian,
                "__label__te": Languages.telugu,
                "__label__kk": Languages.kazakh,
                "__label__war": Languages.waray,
                "__label__lt": Languages.lithuanian,
                "__label__gl": Languages.scottish,
                "__label__sk": Languages.slovak,
                "__label__bn": Languages.benin,
                "__label__eu": Languages.basque,
                "__label__sl": Languages.slovenian,
                "__label__kn": Languages.not_defined,
                "__label__ml": Languages.malayalam,
                "__label__mr": Languages.marathi,
                "__label__et": Languages.estonian,
                "__label__az": Languages.azerbaijani,
                "__label__ms": Languages.not_defined,
                "__label__sq": Languages.albanian,
                "__label__la": Languages.latin,
                "__label__bs": Languages.bosnian,
                "__label__nn": Languages.norwegian_nynorsk,
                "__label__ur": Languages.urdu,
                "__label__lv": Languages.not_defined,
                "__label__my": Languages.not_defined,
                "__label__tt": Languages.not_defined,
                "__label__af": Languages.not_defined,
                "__label__oc": Languages.not_defined,
                "__label__nds": Languages.not_defined,
                "__label__ky": Languages.not_defined,
                "__label__ast": Languages.not_defined,
                "__label__tl": Languages.not_defined,
                "__label__is": Languages.not_defined,
                "__label__ia": Languages.not_defined,
                "__label__si": Languages.not_defined,
                "__label__gu": Languages.not_defined,
                "__label__km": Languages.not_defined,
                "__label__br": Languages.not_defined,
                "__label__ba": Languages.not_defined,
                "__label__uz": Languages.not_defined,
                "__label__bo": Languages.not_defined,
                "__label__pa": Languages.not_defined,
                "__label__vo": Languages.not_defined,
                "__label__als": Languages.not_defined,
                "__label__ne": Languages.not_defined,
                "__label__cy": Languages.not_defined,
                "__label__jbo": Languages.not_defined,
                "__label__fy": Languages.not_defined,
                "__label__mn": Languages.not_defined,
                "__label__lb": Languages.not_defined,
                "__label__ce": Languages.not_defined,
                "__label__ug": Languages.not_defined,
                "__label__tg": Languages.not_defined,
                "__label__sco": Languages.not_defined,
                "__label__sa": Languages.not_defined,
                "__label__cv": Languages.not_defined,
                "__label__jv": Languages.not_defined,
                "__label__min": Languages.not_defined,
                "__label__io": Languages.not_defined,
                "__label__or": Languages.not_defined,
                "__label__as": Languages.not_defined,
                "__label__new": Languages.not_defined,
                "__label__ga": Languages.not_defined,
                "__label__mg": Languages.not_defined,
                "__label__an": Languages.not_defined,
                "__label__ckb": Languages.not_defined,
                "__label__sw": Languages.not_defined,
                "__label__bar": Languages.not_defined,
                "__label__lmo": Languages.not_defined,
                "__label__yi": Languages.not_defined,
                "__label__arz": Languages.not_defined,
                "__label__mhr": Languages.not_defined,
                "__label__azb": Languages.not_defined,
                "__label__sah": Languages.not_defined,
                "__label__pnb": Languages.not_defined,
                "__label__su": Languages.not_defined,
                "__label__bpy": Languages.not_defined,
                "__label__pms": Languages.not_defined,
                "__label__ilo": Languages.not_defined,
                "__label__wuu": Languages.not_defined,
                "__label__ku": Languages.not_defined,
                "__label__ps": Languages.not_defined,
                "__label__ie": Languages.not_defined,
                "__label__xmf": Languages.not_defined,
                "__label__yue": Languages.not_defined,
                "__label__gom": Languages.not_defined,
                "__label__li": Languages.not_defined,
                "__label__mwl": Languages.not_defined,
                "__label__kw": Languages.not_defined,
                "__label__sd": Languages.not_defined,
                "__label__hsb": Languages.not_defined,
                "__label__scn": Languages.not_defined,
                "__label__gd": Languages.not_defined,
                "__label__pam": Languages.not_defined,
                "__label__bh": Languages.not_defined,
                "__label__mai": Languages.not_defined,
                "__label__vec": Languages.not_defined,
                "__label__mt": Languages.not_defined,
                "__label__dv": Languages.not_defined,
                "__label__wa": Languages.not_defined,
                "__label__mzn": Languages.not_defined,
                "__label__am": Languages.not_defined,
                "__label__qu": Languages.not_defined,
                "__label__eml": Languages.not_defined,
                "__label__cbk": Languages.not_defined,
                "__label__tk": Languages.not_defined,
                "__label__rm": Languages.not_defined,
                "__label__os": Languages.not_defined,
                "__label__vls": Languages.not_defined,
                "__label__yo": Languages.not_defined,
                "__label__lo": Languages.not_defined,
                "__label__lez": Languages.not_defined,
                "__label__so": Languages.not_defined,
                "__label__myv": Languages.not_defined,
                "__label__diq": Languages.not_defined,
                "__label__mrj": Languages.not_defined,
                "__label__dsb": Languages.not_defined,
                "__label__frr": Languages.not_defined,
                "__label__ht": Languages.not_defined,
                "__label__gn": Languages.not_defined,
                "__label__bxr": Languages.not_defined,
                "__label__kv": Languages.not_defined,
                "__label__sc": Languages.not_defined,
                "__label__nah": Languages.not_defined,
                "__label__krc": Languages.not_defined,
                "__label__bcl": Languages.not_defined,
                "__label__nap": Languages.not_defined,
                "__label__gv": Languages.not_defined,
                "__label__av": Languages.not_defined,
                "__label__rue": Languages.not_defined,
                "__label__xal": Languages.not_defined,
                "__label__pfl": Languages.not_defined,
                "__label__dty": Languages.not_defined,
                "__label__hif": Languages.not_defined,
                "__label__co": Languages.not_defined,
                "__label__lrc": Languages.not_defined,
                "__label__vep": Languages.not_defined,
                "__label__tyv": Languages.not_defined,
            },
            model_wrapper="FasttextLangDetector",
        ),
    }

    @staticmethod
    def get_full_path_weights(name: str) -> str:
        """
        Returns the absolute path of weights.

        Note, that weights are sometimes not defined by only one artefact. The returned string will only represent one
        weights artefact.

        :param name: model name
        :return: absolute weight path
        """
        profile = ModelCatalog.get_profile(name)
        if profile.name:
            return os.path.join(get_weights_dir_path(), profile.name)
        log_once(
            f"Model {name} is not registered. Please make sure the weights are available in the weights cache "
            f"directory or the full path you provide is correct"
        )
        if os.path.isfile(name):
            return name
        return os.path.join(get_weights_dir_path(), name)

    @staticmethod
    def get_full_path_configs(name: str) -> str:
        """
        Return the absolute path of configs for some given weights. Alternatively, pass last a path to a config file
        (without the base path to the cache config directory).

        Note, that configs are sometimes not defined by only one file. The returned string will only represent one
        file.

        :param name: model name
        :return: absolute path to the config
        """
        profile = ModelCatalog.get_profile(name)
        if profile.config is not None:
            return os.path.join(get_configs_dir_path(), profile.config)
        return os.path.join(get_configs_dir_path(), name)

    @staticmethod
    def get_full_path_preprocessor_configs(name: str) -> str:
        """
        Return the absolute path of preprocessor configs for some given weights. Preprocessor are occasionally provided
        by the transformer library.

        :param name: model name
        :return: absolute path to the preprocessor config
        """
        profile = ModelCatalog.get_profile(name)
        if profile.preprocessor_config is not None:
            return os.path.join(get_configs_dir_path(), profile.preprocessor_config)
        return os.path.join(get_configs_dir_path(), name)

    @staticmethod
    def get_model_list() -> List[str]:
        """
        Returns a list of absolute paths of registered models.
        """
        return [os.path.join(get_weights_dir_path(), profile.name) for profile in ModelCatalog.CATALOG.values()]

    @staticmethod
    def is_registered(path_weights: str) -> bool:
        """
        Checks if some weights belong to a registered model

        :param path_weights: relative or absolute path
        :return: True if the weights are registered in `ModelCatalog`
        """
        if (ModelCatalog.get_full_path_weights(path_weights) in ModelCatalog.get_model_list()) or (
            path_weights in ModelCatalog.get_model_list()
        ):
            return True
        return False

    @staticmethod
    def get_profile(name: str) -> ModelProfile:
        """
        Returns the profile of given model name, i.e. the config file, size and urls.

        :param name: model name
        :return: A dict of model/weights profiles
        """
        profile = ModelCatalog.CATALOG.get(name)
        if profile is not None:
            return copy(profile)
        return ModelProfile(name="", description="", size=[0], tp_model=False)

    @staticmethod
    def register(name: str, profile: ModelProfile) -> None:
        """
        Register a model with its profile

        :param name: Name of the model. We use the file name of the model along with its path (starting from the
                     weights .cache dir. e.g. 'my_model/model_123.pkl'.
        :param profile: profile of the model
        """
        if name in ModelCatalog.CATALOG:
            raise KeyError("Model already registered")
        ModelCatalog.CATALOG[name] = profile


def get_tp_weight_names(name: str) -> List[str]:
    """
    Given a path to some model weights it will return all file names according to TP naming convention

    :param name: TP model name
    :return: A list of TP file names
    """
    _, file_name = os.path.split(name)
    prefix, _ = file_name.split(".")
    weight_names = []
    for suffix in ["data-00000-of-00001", "index"]:
        weight_names.append(prefix + "." + suffix)

    return weight_names


def print_model_infos(add_description: bool = True, add_config: bool = True, add_categories: bool = True) -> None:
    """
    Prints a table with all registered model profiles and some of their attributes (name, description, config and
    categories)
    """

    profiles = ModelCatalog.CATALOG.values()
    num_columns = min(6, len(profiles))
    infos = []
    for profile in profiles:
        tbl_input: List[Union[Mapping[str, ObjectTypes], str]] = [profile.name]
        if add_description:
            tbl_input.append(profile.description)
        if add_config:
            tbl_input.append(profile.config if profile.config else "")
        if add_categories:
            tbl_input.append(profile.categories if profile.categories else {})
        infos.append(tbl_input)
    tbl_header = ["name"]
    if add_description:
        tbl_header.append("description")
    if add_config:
        tbl_header.append("config")
    if add_categories:
        tbl_header.append("categories")
    table = tabulate(
        infos,
        headers=tbl_header * (num_columns // 2),
        tablefmt="fancy_grid",
        stralign="left",
        numalign="left",
    )
    print(colored(table, "cyan"))


class ModelDownloadManager:
    """
    Class for organizing downloads of config files and weights from various sources. Internally, it will use model
    profiles to know where things are stored.

        # if you are not sure about the model name use the ModelCatalog
        ModelDownloadManager.maybe_download_weights_and_configs("layout/model-800000_inf_only.data-00000-of-00001")
    """

    @staticmethod
    def maybe_download_weights_and_configs(name: str) -> str:
        """
        Check if some model is registered. If yes, it will check if their weights
        must be downloaded. Only weights that have not the same expected size will be downloaded again.

        :param name: A path to some model weights
        :return: Absolute path to model weights if model is registered
        """

        absolute_path_weights = ModelCatalog.get_full_path_weights(name)
        file_names: List[str] = []
        if ModelCatalog.is_registered(name):
            profile = ModelCatalog.get_profile(name)
            # there is nothing to download if hf_repo_id or urls is not provided
            if not profile.hf_repo_id and not profile.urls:
                return absolute_path_weights
            # determine the right model name
            if profile.tp_model:
                file_names = get_tp_weight_names(name)
            else:
                model_name = profile.hf_model_name
                if model_name is None:
                    # second try. Check if a url is provided
                    if profile.urls is None:
                        raise ValueError("hf_model_name and urls cannot be both None")
                    for url in profile.urls:
                        file_names.append(url.split("/")[-1])
                else:
                    file_names.append(model_name)
            if profile.hf_repo_id:
                ModelDownloadManager.load_model_from_hf_hub(profile, absolute_path_weights, file_names)
                absolute_path_configs = ModelCatalog.get_full_path_configs(name)
                ModelDownloadManager.load_configs_from_hf_hub(profile, absolute_path_configs)
            else:
                ModelDownloadManager._load_from_gd(profile, absolute_path_weights, file_names)

            return absolute_path_weights

        return absolute_path_weights

    @staticmethod
    def load_model_from_hf_hub(profile: ModelProfile, absolute_path: str, file_names: List[str]) -> None:
        """
        Load a model from the Huggingface hub for a given profile and saves the model at the directory of the given
        path.

        :param profile: Profile according to `ModelCatalog.get_profile(path_weights)`
        :param absolute_path: Absolute path (incl. file name) of target file
        :param file_names: Optionally, replace the file name of the ModelCatalog. This is necessary e.g. for Tensorpack
                           models
        """
        repo_id = profile.hf_repo_id
        if repo_id is None:
            raise ValueError("hf_repo_id cannot be None")
        directory, _ = os.path.split(absolute_path)

        for expect_size, file_name in zip(profile.size, file_names):
            size = ModelDownloadManager._load_from_hf_hub(repo_id, file_name, directory)
            if expect_size is not None and size != expect_size:
                logger.error(
                    "File downloaded from %s does not match the expected size! You may have downloaded"
                    " a broken file, or the upstream may have modified the file.",
                    repo_id,
                )

    @staticmethod
    def _load_from_gd(profile: ModelProfile, absolute_path: str, file_names: List[str]) -> None:
        if profile.urls is None:
            raise ValueError("urls cannot be None")
        for size, url, file_name in zip(profile.size, profile.urls, file_names):
            directory, _ = os.path.split(absolute_path)
            download(str(url), directory, file_name, int(size))

    @staticmethod
    def load_configs_from_hf_hub(profile: ModelProfile, absolute_path: str) -> None:
        """
        Load config file(s) from the Huggingface hub for a given profile and saves the model at the directory of the
        given path.

        :param profile: Profile according to `ModelCatalog.get_profile(path_weights)`
        :param absolute_path:  Absolute path (incl. file name) of target file
        """

        repo_id = profile.hf_repo_id
        if repo_id is None:
            raise ValueError("hf_repo_id cannot be None")
        directory, _ = os.path.split(absolute_path)
        if not profile.hf_config_file:
            raise ValueError("hf_config_file cannot be None")
        for file_name in profile.hf_config_file:
            ModelDownloadManager._load_from_hf_hub(repo_id, file_name, directory)

    @staticmethod
    def _load_from_hf_hub(repo_id: str, file_name: str, cache_directory: str, force_download: bool = False) -> int:
        url = hf_hub_url(repo_id=repo_id, filename=file_name)
        token = os.environ.get("HF_CREDENTIALS")
        f_path = cached_download(
            url,
            cache_dir=cache_directory,
            force_filename=file_name,
            force_download=force_download,
            token=token,
            legacy_cache_layout=True,
        )
        if f_path:
            stat_info = os.stat(f_path)
            size = stat_info.st_size

            assert size > 0, f"Downloaded an empty file from {url}!"
            return size
        raise TypeError("Returned value from cached_download cannot be Null")



# -*- coding: utf-8 -*-
# File: text.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for text extraction pipeline component
"""
from copy import copy, deepcopy
from itertools import chain
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union

from ..datapoint.annotation import ImageAnnotation
from ..datapoint.image import Image
from ..extern.base import ObjectDetector, PdfMiner, TextRecognizer
from ..extern.tessocr import TesseractOcrDetector
from ..utils.detection_types import ImageType, JsonDict
from ..utils.logger import logger
from ..utils.settings import LayoutType, PageType, Relationships, TypeOrStr, WordType, get_type
from .base import PipelineComponent, PredictorPipelineComponent
from .registry import pipeline_component_registry

__all__ = ["TextExtractionService", "TextOrderService"]


@pipeline_component_registry.register("TextExtractionService")
class TextExtractionService(PredictorPipelineComponent):
    """
    Pipeline component for extracting text. Any detector can be selected, provided that it can evaluate a
    numpy array as an image.

    Text extraction can either be carried out over the entire image or over selected regions of interests (ROIs).
    ROIs are layout components that have been determined by means of a pipeline component that has been run through
    beforehand. ROI extraction is particularly suitable when an OCR component is selected as the detector and the
    document has a complex structure. Instead of transferring the entire image, only the ROIs are transferred to
    the detector. Since the ROI has a simpler structure than the entire document page, it can significantly improve
    the OCR results.

    Text components (currently only words) are attached to the image as image annotations. A relation is assigned in
    relation to text and ROI or in relation to text and the entire image. When selecting ROIs, only the selected
    categories are taken into account during processing. ROIs that are not selected are not presented to the
    detector.

        textract_predictor = TextractOcrDetector()
        text_extract = TextExtractionService(textract_predictor)

        pipe = DoctectionPipe([text_extract])
        df = pipe.analyze(path="path/to/document.pdf")

        for dp in df:
            ...
    """

    def __init__(
        self,
        text_extract_detector: Union[ObjectDetector, PdfMiner, TextRecognizer],
        extract_from_roi: Optional[Union[Sequence[TypeOrStr], TypeOrStr]] = None,
        run_time_ocr_language_selection: bool = False,
    ):
        """
        :param text_extract_detector: ObjectDetector
        :param extract_from_roi: one or more category names for roi selection
        """

        if extract_from_roi is None:
            extract_from_roi = []
        self.extract_from_category = (
            [get_type(extract_from_roi)]
            if isinstance(extract_from_roi, str)
            else [get_type(roi_category) for roi_category in extract_from_roi]
        )
        super().__init__(self._get_name(text_extract_detector.name), text_extract_detector)
        if self.extract_from_category:
            if not isinstance(self.predictor, (ObjectDetector, TextRecognizer)):
                raise TypeError("Predicting from a cropped image requires to pass an ObjectDetector or TextRecognizer.")
        if run_time_ocr_language_selection:
            assert isinstance(self.predictor, TesseractOcrDetector), (
                "Only TesseractOcrDetector supports multiple " "languages"
            )

        self.run_time_ocr_language_selection = run_time_ocr_language_selection

    def serve(self, dp: Image) -> None:
        maybe_batched_text_rois = self.get_text_rois(dp)
        for text_roi in maybe_batched_text_rois:
            ann_id = None
            if isinstance(text_roi, ImageAnnotation):
                ann_id = text_roi.annotation_id
            predictor_input = self.get_predictor_input(text_roi)
            if predictor_input is None:
                raise ValueError("predictor_input cannot be None")
            width, height = None, None
            if self.run_time_ocr_language_selection:
                self.predictor.set_language(dp.summary.get_sub_category(PageType.language).value)  # type: ignore
            detect_result_list = self.predictor.predict(predictor_input)  # type: ignore
            if isinstance(self.predictor, PdfMiner):
                width, height = self.predictor.get_width_height(predictor_input)  # type: ignore

            for detect_result in detect_result_list:
                if isinstance(self.predictor, TextRecognizer):
                    detect_ann_id = detect_result.uuid
                else:
                    detect_ann_id = self.dp_manager.set_image_annotation(
                        detect_result, ann_id, True, detect_result_max_width=width, detect_result_max_height=height
                    )
                if detect_ann_id is not None:
                    self.dp_manager.set_container_annotation(
                        WordType.characters,
                        None,
                        WordType.characters,
                        detect_ann_id,
                        detect_result.text if detect_result.text is not None else "",
                        detect_result.score,
                    )
                    if detect_result.block:
                        self.dp_manager.set_category_annotation(
                            WordType.block, detect_result.block, WordType.block, detect_ann_id
                        )
                    if detect_result.line:
                        self.dp_manager.set_category_annotation(
                            WordType.text_line, detect_result.line, WordType.text_line, detect_ann_id
                        )

    def get_text_rois(self, dp: Image) -> Sequence[Union[Image, ImageAnnotation, List[ImageAnnotation]]]:
        """
        Return image rois based on selected categories. As this selection makes only sense for specific text extractors
        (e.g. those who do proper OCR and do not mine from text from native pdfs) it will do some sanity checks.
        :return: list of ImageAnnotation or Image
        """

        if self.extract_from_category:
            if self.predictor.accepts_batch:
                return [dp.get_annotation(category_names=self.extract_from_category)]
            return dp.get_annotation(category_names=self.extract_from_category)
        return [dp]

    def get_predictor_input(
        self, text_roi: Union[Image, ImageAnnotation, List[ImageAnnotation]]
    ) -> Optional[Union[bytes, ImageType, List[Tuple[str, ImageType]]]]:
        """
        Return raw input for a given `text_roi`. This can be a numpy array or pdf bytes and depends on the chosen
        predictor.

        :param text_roi: `Image` or `ImageAnnotation`
        :return: pdf bytes or numpy array
        """

        if isinstance(text_roi, ImageAnnotation):
            if text_roi.image is None:
                raise ValueError("text_roi.image cannot be None")
            if text_roi.image.image is None:
                raise ValueError("text_roi.image.image cannot be None")
            return text_roi.image.image
        if isinstance(self.predictor, ObjectDetector):
            if not isinstance(text_roi, Image):
                raise ValueError("text_roi must be an image")
            return text_roi.image
        if isinstance(text_roi, list):
            assert all(roi.image is not None for roi in text_roi)
            assert all(roi.image.image is not None for roi in text_roi)  # type: ignore
            return [(roi.annotation_id, roi.image.image) for roi in text_roi]  # type: ignore
        return text_roi.pdf_bytes

    def get_meta_annotation(self) -> JsonDict:
        if self.extract_from_category:
            sub_cat_dict = {category: {WordType.characters} for category in self.extract_from_category}
        else:
            if not isinstance(self.predictor, (ObjectDetector, PdfMiner)):
                raise TypeError(
                    f"self.predictor must be of type ObjectDetector or PdfMiner but is of type "
                    f"{type(self.predictor)}"
                )
            sub_cat_dict = {category: {WordType.characters} for category in self.predictor.possible_categories()}
        return dict(
            [
                (
                    "image_annotations",
                    self.predictor.possible_categories()
                    if isinstance(self.predictor, (ObjectDetector, PdfMiner))
                    else [],
                ),
                ("sub_categories", sub_cat_dict),
                ("relationships", {}),
                ("summaries", []),
            ]
        )

    @staticmethod
    def _get_name(text_detector_name: str) -> str:
        return f"text_extract_{text_detector_name}"

    def clone(self) -> "PredictorPipelineComponent":
        predictor = self.predictor.clone()
        if not isinstance(predictor, (ObjectDetector, PdfMiner, TextRecognizer)):
            raise ValueError(f"predictor must be of type ObjectDetector or PdfMiner, but is of type {type(predictor)}")
        return self.__class__(predictor, deepcopy(self.extract_from_category), self.run_time_ocr_language_selection)


def _reading_lines(image_id: str, word_anns: List[ImageAnnotation]) -> List[Tuple[int, str]]:
    reading_lines = []
    rows: List[Dict[str, float]] = []
    for word in word_anns:
        if word.image is not None:
            bounding_box = word.image.get_embedding(image_id)
        else:
            bounding_box = word.bounding_box
        row_found = False
        for idx, row in enumerate(rows):
            row_cy = (row["upper"] + row["lower"]) / 2

            if (row["upper"] < bounding_box.cy < row["lower"]) or (bounding_box.uly < row_cy < bounding_box.lry):
                reading_lines.append((idx, word.annotation_id, bounding_box.cx))
                row_found = True
                break

        if not row_found:
            rows.append({"upper": bounding_box.uly, "lower": bounding_box.lry})
            reading_lines.append((len(rows) - 1, word.annotation_id, bounding_box.cx))

    rows_dict = {k: rows[k] for k in range(len(rows))}
    rows_dict = {
        idx: key[0] for idx, key in enumerate(sorted(rows_dict.items(), key=lambda it: it[1]["upper"]))  # type:ignore
    }
    reading_lines.sort(key=lambda x: (rows_dict[x[0]], x[2]))
    return [(idx + 1, word[1]) for idx, word in enumerate(reading_lines)]


def _reading_columns(
    dp: Image,
    anns: List[ImageAnnotation],
    starting_point_tolerance: float = 0.01,
    height_tolerance: float = 3.0,
) -> List[Tuple[int, str]]:
    reading_blocks = []
    columns: List[Dict[str, float]] = []
    anns.sort(key=lambda x: (x.bounding_box.cy, x.bounding_box.cx))  # type: ignore
    for ann in anns:
        if ann.image is not None:
            bounding_box = ann.image.get_embedding(dp.image_id)
        else:
            bounding_box = ann.bounding_box

        if bounding_box.absolute_coords:
            rel_coords_box = bounding_box.transform(dp.width, dp.height)
        else:
            rel_coords_box = bounding_box

        column_found = False
        for idx, col in enumerate(columns):
            # if the x-coordinate left and right is within starting_point_tolerance (first_condition and
            # second_condition) or if x-coordinate left and right is within the left or right border of the column
            # then the annotation will belong to this column and column left/right will be re-adjusted

            first_condition = all(
                (
                    col["left"] - starting_point_tolerance < rel_coords_box.ulx,
                    rel_coords_box.lrx < col["right"] + starting_point_tolerance,
                )
            )
            second_condition = all(
                (
                    rel_coords_box.ulx - starting_point_tolerance < col["left"],
                    col["right"] < rel_coords_box.lrx + starting_point_tolerance,
                )
            )

            third_condition = abs(rel_coords_box.uly - col["bottom"]) < height_tolerance * rel_coords_box.height
            fourth_condition = abs(rel_coords_box.lry - col["top"]) < height_tolerance * rel_coords_box.height

            if (first_condition and (third_condition or fourth_condition)) or (  # pylint: disable=R0916
                second_condition and (third_condition or fourth_condition)
            ):
                reading_blocks.append((idx, ann.annotation_id))
                # update the top and right with the new line added.
                col["left"] = min(rel_coords_box.ulx, col["left"])
                col["top"] = min(rel_coords_box.uly, col["top"])
                col["right"] = max(rel_coords_box.lrx, col["right"])
                col["bottom"] = max(rel_coords_box.lry, col["bottom"])
                column_found = True
                break

        if not column_found:
            columns.append(
                {
                    "left": rel_coords_box.ulx,
                    "right": rel_coords_box.lrx,
                    "top": rel_coords_box.uly,
                    "bottom": rel_coords_box.lry,
                }
            )
            # update the top and right with the new reading block added.
            reading_blocks.append((len(columns) - 1, ann.annotation_id))

    # building connected components of columns
    connected_components: List[Dict[str, Any]] = []
    for idx, col in enumerate(columns):
        col["id"] = idx
        component_found = False
        for comp in connected_components:
            if comp["top"] < col["top"] < comp["bottom"] or comp["top"] < col["bottom"] < comp["bottom"]:
                comp["top"] = min(comp["top"], col["top"])
                comp["bottom"] = max(comp["bottom"], col["bottom"])
                comp["left"] = col["left"]
                comp["column"].append(col)
                component_found = True
                break
        if not component_found:
            connected_components.append(
                {"top": col["top"], "bottom": col["bottom"], "left": col["left"], "column": [col]}
            )

    # next, sorting columns in connected components by increasing x-value
    for comp in connected_components:
        comp["column"].sort(key=lambda x: x["left"])

    # finally, sorting connected components by increasing y-value
    connected_components.sort(key=lambda x: x["top"])
    columns = list(chain(*[comp["column"] for comp in connected_components]))

    # old to new mapping
    columns_dict = {k: col["id"] for k, col in enumerate(columns)}
    _blocks = [(columns_dict[x[0]], x[1]) for x in reading_blocks]
    _blocks.sort(key=lambda x: x[0])
    reading_blocks = [(idx + 1, block[1]) for idx, block in enumerate(_blocks)]
    return reading_blocks


@pipeline_component_registry.register("TextOrderService")
class TextOrderService(PipelineComponent):
    """
    Reading order of words within floating text blocks as well as reading order of blocks within simple text blocks.
    To understand the difference between floating text blocks and simple text blocks consider a page containing an
    article and a table. Table cells are text blocks that contain words which must be sorted.
    However, they do not belong to floating text that encircle a table. They are rather an element that is supposed to
    be read independently.

    A heuristic argument for its ordering is used where the underlying assumption is the reading order from left to
    right.

        - For the reading order within a text block, text containers (i.e. image annotations that contain character
          sub annotations) are sorted based on their bounding box center and then lines are formed: Each word induces a
          new line, provided that its center is not in a line that has already
          been created by an already processed word. The entire block width is defined as the line width
          and the upper or lower line limit of the word bounding box as the upper or lower line limit. The reading order
          of the words is from left to right within a line. The reading order of the lines is from top to bottom.

        - For the reading order of text blocks within a page, the blocks are sorted using a similar procedure, with the
          difference that columns are formed instead of lines. Column lengths are defined as the length of the entire
          page and the left and right text block boundaries as the left and right column boundaries.

    A category annotation per word is generated, which fixes the order per word in the block, as well as a category
    annotation per block, which saves the reading order of the block per page.

    The blocks are defined in `_floating_text_block_names` and text blocks in `_floating_text_block_names`.

        order = TextOrderService(text_container=names.C.WORD,
                                 floating_text_block_names=[names.C.TITLE, names.C.TEXT, names.C.LIST],
                                 text_block_names=[names.C.TITLE, names.C.TEXT, names.C.LIST, names.C.CELL,
                                                   names.C.HEAD, names.C.BODY])
    """

    def __init__(
        self,
        text_container: str,
        floating_text_block_names: Optional[Union[str, Sequence[str]]] = None,
        text_block_names: Optional[Union[str, Sequence[str]]] = None,
        text_containers_to_text_block: bool = False,
    ) -> None:
        """
        :param text_container: name of an image annotation that has a CHARS sub category. These annotations will be
                               ordered within all text blocks.
        :param floating_text_block_names: name of image annotation that belong to floating text. These annotations form
                                          the highest hierarchy of text blocks that will ordered to generate a sensible
                                          output of text
        :param text_block_names: name of image annotation that have a relation with text containers (or which might be
                                 text containers themselves).
        :param text_containers_to_text_block: Text containers are in general no text blocks and belong to a lower
                                              hierarchy. However, if a text container is not assigned to a text block
                                              you can add it to the text block ordering to ensure that the full text is
                                              part of the subsequent sub process. Note however, that if the text
                                              container is on word level rather than line level, the results will not be
                                              very convincing
        """
        if isinstance(floating_text_block_names, str):
            floating_text_block_names = [floating_text_block_names]
        elif floating_text_block_names is None:
            floating_text_block_names = []
        if isinstance(text_block_names, str):
            text_block_names = [text_block_names]
        elif text_block_names is None:
            text_block_names = []

        self._text_container = text_container
        self._floating_text_block_names = floating_text_block_names
        self._text_block_names = text_block_names
        self._text_containers_to_text_block = text_containers_to_text_block
        self._init_sanity_checks()
        super().__init__("text_order")

    def serve(self, dp: Image) -> None:
        # select all text blocks that are considered to be relevant for page text. This does not include some layout
        # items that have to be considered independently (e.g. tables). Order the blocks by column wise reading order
        floating_text_block_anns = dp.get_annotation(category_names=self._floating_text_block_names)
        number_floating_text_block_anns_orig = len(floating_text_block_anns)
        # maybe add all text containers that are not mapped to any text block
        if self._text_containers_to_text_block:
            text_block_anns = dp.get_annotation(category_names=self._text_block_names)
            text_ann_ids = list(
                chain(*[text_block.get_relationship(Relationships.child) for text_block in text_block_anns])
            )
            text_container_anns = dp.get_annotation(category_names=self._text_container)
            text_container_anns = [ann for ann in text_container_anns if ann.annotation_id not in text_ann_ids]
            floating_text_block_anns.extend(text_container_anns)

        # estimating reading columns. We will only do this if we have some text blocks that are no text_containers
        # (number_text_block_anns_orig >0) or if the text container is not a word. Otherwise, we will have to skip that
        # part
        if self._text_container != LayoutType.word or number_floating_text_block_anns_orig:
            raw_reading_order_list = _reading_columns(dp, floating_text_block_anns, 0.05, 2.0)

            for raw_reading_order in raw_reading_order_list:
                self.dp_manager.set_category_annotation(
                    Relationships.reading_order, raw_reading_order[0], Relationships.reading_order, raw_reading_order[1]
                )

        # next we select all blocks that might contain text. We sort all text within these blocks
        block_anns = dp.get_annotation(category_names=self._text_block_names)
        for text_block in block_anns:
            text_container_ann_ids = text_block.get_relationship(Relationships.child)
            text_container_anns = dp.get_annotation(
                annotation_ids=text_container_ann_ids,
                category_names=self._text_container,
            )
            raw_reading_order_list = _reading_lines(dp.image_id, text_container_anns)
            for raw_reading_order in raw_reading_order_list:
                self.dp_manager.set_category_annotation(
                    Relationships.reading_order, raw_reading_order[0], Relationships.reading_order, raw_reading_order[1]
                )

        # this is the setting where we order words without having text blocks
        if not block_anns:
            text_container_anns = dp.get_annotation(category_names=self._text_container)
            # some OCR systems return textline and blocks. If they are available we will sort first by block, then by
            # line and finally by center x coord.
            if text_container_anns:
                text_container_ann = text_container_anns[0]
                if WordType.block and WordType.text_line in text_container_ann.sub_categories:
                    text_container_position = [
                        (
                            int(ann.get_sub_category(WordType.block).category_id),
                            int(ann.get_sub_category(WordType.text_line).category_id),
                            ann.bounding_box.cx,  # type: ignore
                            ann.annotation_id,
                        )
                        for ann in text_container_anns
                    ]
                    text_container_position.sort(key=lambda x: (x[0], x[1], x[2]))
                    for position, element in enumerate(text_container_position):
                        self.dp_manager.set_category_annotation(
                            Relationships.reading_order, position, Relationships.reading_order, element[3]
                        )
                else:
                    # Last try. We only form lines without and define a reading from this
                    raw_reading_order_list = _reading_lines(dp.image_id, text_container_anns)
                    for raw_reading_order in raw_reading_order_list:
                        self.dp_manager.set_category_annotation(
                            Relationships.reading_order,
                            raw_reading_order[0],
                            Relationships.reading_order,
                            raw_reading_order[1],
                        )

    def clone(self) -> PipelineComponent:
        return self.__class__(
            copy(self._text_container),
            deepcopy(self._floating_text_block_names),
            deepcopy(self._text_block_names),
            deepcopy(self._text_containers_to_text_block),
        )

    def _init_sanity_checks(self) -> None:
        assert self._text_container in [LayoutType.word, LayoutType.line], (
            f"text_container must be either {LayoutType.word} or " f"{LayoutType.line}"
        )
        assert set(self._floating_text_block_names) <= set(
            self._text_block_names
        ), "floating_text_block_names must be a subset of text_block_names"
        if (
            self._text_container == LayoutType.word
            and self._text_containers_to_text_block
            and not self._floating_text_block_names
        ):
            logger.info(
                "Choosing %s text_container while choosing no text_blocks will give no sensible "
                "results. Choose %s text_container if you do not have text_blocks available.",
                LayoutType.word,
                LayoutType.line,
            )

    def get_meta_annotation(self) -> JsonDict:
        anns_with_reading_order = list(deepcopy(self._floating_text_block_names))
        anns_with_reading_order.extend([LayoutType.word, LayoutType.line])
        return dict(
            [
                ("image_annotations", []),
                ("sub_categories", {category: {Relationships.reading_order} for category in anns_with_reading_order}),
                ("relationships", {}),
                ("summaries", []),
            ]
        )



# -*- coding: utf-8 -*-
# File: test_registry.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing the module eval.registry
"""

from typing import Any, List, Tuple
from unittest.mock import MagicMock

from pytest import mark

from deepdoctection.dataflow import DataFlow
from deepdoctection.datasets.info import DatasetCategories
from deepdoctection.eval import metric_registry
from deepdoctection.eval.base import MetricBase
from deepdoctection.utils.detection_types import JsonDict
from deepdoctection.utils.file_utils import apted_available


@mark.basic
def test_metric_registry_has_all_build_in_metric_registered() -> None:
    """
    test metric registry has all metrics registered
    """
    if apted_available():
        # cannot register when apted not installed
        assert len(metric_registry.get_all()) == 10
    else:
        assert len(metric_registry.get_all()) == 9


@mark.basic
def test_metric_registry_registered_new_metric() -> None:
    """
    test, that the new generated metric "TestMetric" can be registered and retrieved from registry
    """

    @metric_registry.register("testmetric")
    class TestMetric(MetricBase):
        """
        TestMetric
        """

        @classmethod
        def get_distance(
            cls,
            dataflow_gt: DataFlow,
            dataflow_predictions: DataFlow,
            categories: DatasetCategories,
        ) -> List[JsonDict]:
            """
            get distance
            """
            return MagicMock()

        @classmethod
        def dump(
            cls, dataflow_gt: DataFlow, dataflow_predictions: DataFlow, categories: DatasetCategories
        ) -> Tuple[Any, Any]:
            """
            dump
            """
            return MagicMock()

    # Act
    test = metric_registry.get("testmetric")

    # Assert
    assert test == TestMetric



# -*- coding: utf-8 -*-
# File: hfstruct.py

# Copyright 2023 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for mapping annotations into standard Huggingface Detr input structure for training
"""

import os
from dataclasses import dataclass, field
from typing import Dict, List, Literal, Optional, Sequence, Union

import numpy as np
from transformers import BatchFeature, DetrFeatureExtractor

from ..datapoint.image import Image
from ..mapper.maputils import curry
from ..mapper.misc import get_load_image_func
from ..utils.detection_types import JsonDict
from ..utils.settings import ObjectTypes
from ..utils.transform import PadTransform


@curry
def image_to_hf_detr_training(
    dp: Image,
    add_mask: bool = False,
    category_names: Optional[Union[str, ObjectTypes, Sequence[Union[str, ObjectTypes]]]] = None,
) -> Optional[JsonDict]:
    """
    Maps an image to a detr input datapoint dict, that, after collating can be used for training.

    :param dp: Image
    :param add_mask: True is not implemented (yet).
    :param category_names: A list of category names for training a model. Pass nothing to train with all annotations
    :return: Dict with 'image', 'width', 'height', 'image_id', 'annotations' where 'annotations' is a list of dict
             with 'boxes' and 'class_labels'.
    """

    if not os.path.isfile(dp.location) and dp.image is None:
        return None

    output: JsonDict = {"file_name": dp.location}

    if dp.image is not None:
        output["image"] = dp.image.astype("float32")
    output["width"] = dp.width
    output["height"] = dp.height
    output["image_id"] = "".join([c for c in dp.image_id if c.isdigit()])[:8]

    anns = dp.get_annotation(category_names=category_names)

    if not anns:
        return None

    annotations = []

    for ann in anns:
        if ann.image is not None:
            box = ann.image.get_embedding(dp.image_id)
        else:
            box = ann.bounding_box
        if box is None:
            raise ValueError("BoundingBox cannot be None")
        mapped_ann: Dict[str, Union[str, int, List[float]]] = {
            "id": "".join([c for c in ann.annotation_id if c.isdigit()])[:8],
            "image_id": "".join([c for c in dp.image_id if c.isdigit()])[:8],
            "bbox": box.to_list(mode="xywh"),
            "category_id": int(ann.category_id) - 1,
            "area": box.area,
        }
        annotations.append(mapped_ann)

    if add_mask:
        raise NotImplementedError

    output["annotations"] = annotations

    return output


@dataclass
class DetrDataCollator:
    """
    Data collator that will prepare a list of raw features to a BatchFeature that can be used
    to train a Detr or Tabletransformer model.

    :param feature_extractor:  DetrFeatureExtractor
    :param padder: An optional PadTransform instance
    :param return_tensors: "pt" or None
    """

    feature_extractor: DetrFeatureExtractor  # TODO: Replace deprecated DetrFeatureExtractor with DetrImageProcessor
    padder: Optional[PadTransform] = None
    return_tensors: Optional[Literal["pt"]] = field(default="pt")

    def __call__(self, raw_features: List[JsonDict]) -> BatchFeature:
        """
        Creating BatchFeature from a list of dict of raw features.

        :param raw_features: A list of dict with keys: 'image' or 'file_name', "width', "height' and 'annotations'.
                             'annotations' mus be a list of dict as well, where each dict element must contain
                             annotation information following COCO standard.
        :return: BatchFeature
        """
        images_input = []

        for feature in raw_features:
            maybe_image = feature.get("image")
            if maybe_image is None:
                load_func = get_load_image_func(feature["file_name"])
                feature["image"] = load_func(feature["file_name"])
                feature = self.maybe_pad_image_and_transform(feature)
                images_input.append(feature["image"])
            else:
                images_input.append(maybe_image)
        image_features = self.feature_extractor(  # pylint: disable=E1102
            images_input, annotations=raw_features, return_tensors=self.return_tensors
        )

        return image_features

    def maybe_pad_image_and_transform(self, feature: JsonDict) -> JsonDict:
        """
        Pads an 'image' and transforming bounding boxes from annotations.

        :param feature: A dict of raw_features
        :return: Same as input
        """
        if self.padder is None:
            return feature
        feature["image"] = self.padder.apply_image(feature["image"])
        feature["height"] = feature["image"].shape[0]
        feature["width"] = feature["image"].shape[1]
        boxes = np.array([ann["bbox"] for ann in feature["annotations"]])
        box_transform = self.padder.apply_coords(boxes)
        for idx, ann in enumerate(feature["annotations"]):
            ann["bbox"] = box_transform[:, idx].tolist()
        return feature



# -*- coding: utf-8 -*-
# File: custom_serialize.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Methods that convert incoming data to dataflows.
"""

import itertools
import json
import os
from collections import defaultdict
from typing import DefaultDict, Dict, List, Optional, Sequence, Union

from jsonlines import Reader, Writer

from ..utils.context import timed_operation
from ..utils.detection_types import JsonDict, Pathlike
from ..utils.identifier import get_uuid_from_str
from ..utils.pdf_utils import PDFStreamer
from ..utils.tqdm import get_tqdm
from ..utils.utils import FileExtensionError, is_file_extension
from .base import DataFlow
from .common import FlattenData, JoinData, MapData
from .custom import CacheData, CustomDataFromIterable, CustomDataFromList

__all__ = ["SerializerJsonlines", "SerializerFiles", "SerializerCoco", "SerializerPdfDoc", "SerializerTabsepFiles"]


def _reset_df_and_get_length(df: DataFlow) -> int:
    df.reset_state()
    try:
        length = len(df)
    except NotImplementedError:
        length = 0
    return length


class SerializerJsonlines:
    """
    Serialize a dataflow from a jsonlines file. Alternatively, save a dataflow of JSON objects to a .jsonl file.

    **Example:**

            df = SerializerJsonlines.load("path/to/file.jsonl")
            df.reset_state()
            for dp in df:
               ... # is a dict
    """

    @staticmethod
    def load(path: Pathlike, max_datapoints: Optional[int] = None) -> CustomDataFromIterable:
        """
        :param path: a path to a .jsonl file.
        :param max_datapoints: Will stop the iteration once max_datapoints have been streamed

        :return: dataflow to iterate from
        """
        file = open(path, "r")  # pylint: disable=W1514,R1732
        iterator = Reader(file)
        return CustomDataFromIterable(iterator, max_datapoints=max_datapoints)

    @staticmethod
    def save(df: DataFlow, path: Pathlike, file_name: str, max_datapoints: Optional[int] = None) -> None:
        """
        Writes a dataflow iteratively to a .jsonl file. Every datapoint must be a dict where all items are serializable.
        As the length of the dataflow cannot be determined in every case max_datapoint prevents generating an
        unexpectedly large file

        :param df: The dataflow to write from.
        :param path: The path, the .jsonl file to write to.
        :param file_name: name of the target file.
        :param max_datapoints: maximum number of datapoint to consider writing to a file.
        """

        if not os.path.isdir(path):
            raise NotADirectoryError(path)
        if not is_file_extension(file_name, ".jsonl"):
            raise FileExtensionError(f"Expected .jsonl file got {path}")

        df.reset_state()
        with open(os.path.join(path, file_name), "w") as file:  # pylint: disable=W1514
            writer = Writer(file)
            for k, dp in enumerate(df):
                if max_datapoints is None:
                    writer.write(dp)
                elif k < max_datapoints:
                    writer.write(dp)
                else:
                    break


class SerializerTabsepFiles:
    """
    Serialize a dataflow from a tab separated text file. Alternatively, save a dataflow of plain text
    to a .txt file.

    **Example**:

            df = SerializerTabsepFiles.load("path/to/file.txt")

        will yield each text line of the file.
    """

    @staticmethod
    def load(path: Pathlike, max_datapoins: Optional[int] = None) -> CustomDataFromList:
        """
        :param path: a path to a .txt file.
        :param max_datapoins: Will stop the iteration once max_datapoints have been streamed

        :return: dataflow to iterate from
        """

        with open(path, "r", encoding="UTF-8") as file:
            file_list = file.readlines()
        return CustomDataFromList(file_list, max_datapoints=max_datapoins)

    @staticmethod
    def save(df: DataFlow, path: Pathlike, file_name: str, max_datapoints: Optional[int] = None) -> None:
        """
        Writes a dataflow iteratively to a .txt file. Every datapoint must be a string.
        As the length of the dataflow cannot be determined in every case max_datapoint prevents generating an
        unexpectedly large file

        :param df: The dataflow to write from.
        :param path: The path, the .txt file to write to.
        :param file_name: name of the target file.
        :param max_datapoints: maximum number of datapoint to consider writing to a file.
        """

        if not os.path.isdir(path):
            raise NotADirectoryError(path)
        if not is_file_extension(file_name, ".jsonl"):
            raise FileExtensionError(f"Expected .txt file got {path}")

        with open(os.path.join(path, file_name), "w", encoding="UTF-8") as file:
            for k, dp in enumerate(df):
                if max_datapoints is None:
                    file.write(dp)
                elif k < max_datapoints:
                    file.write(dp)
                else:
                    break


class SerializerFiles:
    """
    Serialize files from a directory and all subdirectories. Only one file type can be serialized. Once specified, all
    other types will be filtered out.
    """

    @staticmethod
    def load(
        path: Pathlike,
        file_type: Union[str, Sequence[str]],
        max_datapoints: Optional[int] = None,
        shuffle: Optional[bool] = False,
        sort: Optional[bool] = True,
    ) -> DataFlow:
        """
        Generates a dataflow where a datapoint consists of a string of names of files with respect to some file type.
        If you want to load the files you need to do this in a following step by yourself.

        :param path: A path to some base directory. Will inspect all subdirectories, as well
        :param file_type: A file type (suffix) to look out for (single str or list of stings)
        :param max_datapoints: Stop iteration after passing max_datapoints
        :param shuffle: Shuffle the files, so that the order of appearance in dataflow is random.
        :param sort: If set to "True" it will sort all selected files by its string
        :return: dataflow to iterate from
        """
        df: DataFlow
        df1: DataFlow
        df2: DataFlow
        df3: DataFlow

        if shuffle:
            sort = False
        it1 = os.walk(path, topdown=False)
        it2 = os.walk(path, topdown=False)
        df1 = CustomDataFromIterable(it1)
        df2 = CustomDataFromIterable(it2)
        df1 = MapData(df1, lambda dp: None if len(dp[2]) == 0 else dp)
        df2 = MapData(df2, lambda dp: None if len(dp[2]) == 0 else dp)
        df1 = MapData(df1, lambda dp: [dp[0]] * len(dp[2]))
        df2 = MapData(df2, lambda dp: dp[2])
        df1 = FlattenData(df1)
        df2 = FlattenData(df2)
        df3 = JoinData(df_lists=[df1, df2])
        df3 = MapData(df3, lambda dp: os.path.join(dp[0], dp[1]))
        df = MapData(df3, lambda dp: dp if is_file_extension(dp, file_type) else None)
        if max_datapoints is not None or sort:
            df_list = CacheData(df).get_cache()
            if sort:
                df_list.sort()
            df = CustomDataFromList(df_list, max_datapoints=max_datapoints, shuffle=False)
        elif shuffle:
            df_list = CacheData(df).get_cache()
            df = CustomDataFromList(df_list, shuffle=shuffle)

        return df

    @staticmethod
    def save() -> None:
        """
        Not implemented
        """
        raise NotImplementedError


class CocoParser:
    """
    A simplified version of the Microsoft COCO helper class for reading  annotations. It currently supports only
    bounding box annotations

    :param annotation_file: location of annotation file
    """

    def __init__(self, annotation_file: Optional[Pathlike] = None) -> None:

        self.dataset: JsonDict = {}
        self.anns: Dict[int, JsonDict] = {}
        self.cats: Dict[int, JsonDict] = {}
        self.imgs: Dict[int, JsonDict] = {}

        self.img_to_anns: DefaultDict[int, List[int]] = defaultdict(list)
        self.cat_to_imgs: DefaultDict[int, List[int]] = defaultdict(list)

        if annotation_file is not None:
            with timed_operation(message="Loading annotations to memory"):
                with open(annotation_file, "r", encoding="UTF-8") as file:
                    dataset = json.load(file)
                if not isinstance(dataset, dict):
                    raise TypeError(f"Annotation file format {type(dataset)} for {annotation_file} not supported")
            self.dataset = dataset
            self._create_index()

    def _create_index(self) -> None:
        with timed_operation(message="creating index"):
            anns, cats, imgs = {}, {}, {}
            img_to_anns, cat_to_imgs = defaultdict(list), defaultdict(list)
            if "annotations" in self.dataset:
                for ann in self.dataset["annotations"]:
                    img_to_anns[ann["image_id"]].append(ann)
                    anns[ann["id"]] = ann

            if "images" in self.dataset:
                for img in self.dataset["images"]:
                    imgs[img["id"]] = img

            if "categories" in self.dataset:
                for cat in self.dataset["categories"]:
                    cats[cat["id"]] = cat

            if "annotations" in self.dataset and "categories" in self.dataset:
                for ann in self.dataset["annotations"]:
                    cat_to_imgs[ann["category_id"]].append(ann["image_id"])

            self.anns = anns
            self.img_to_anns = img_to_anns
            self.cat_to_imgs = cat_to_imgs
            self.imgs = imgs
            self.cats = cats

    def info(self) -> None:
        """
        Print information about the annotation file.
        """
        for key, value in self.dataset["info"].items():
            print(f"{key}: {value}")

    def get_ann_ids(
        self,
        img_ids: Optional[Union[int, Sequence[int]]] = None,
        cat_ids: Optional[Union[int, Sequence[int]]] = None,
        area_range: Optional[Sequence[int]] = None,
        is_crowd: Optional[bool] = None,
    ) -> Sequence[int]:
        """
        Get ann ids that satisfy given filter conditions. default skips that filter

        :param img_ids: get anns for given imgs
        :param cat_ids: get anns for given cats
        :param area_range: get anns for given area range (e.g. [0 inf])
        :param is_crowd: get anns for given crowd label (False or True)

        :return: ids: integer array of ann ids
        """

        if img_ids is None:
            img_ids = []
        if cat_ids is None:
            cat_ids = []
        if area_range is None:
            area_range = []

        img_ids = [img_ids] if isinstance(img_ids, int) else img_ids
        cat_ids = [cat_ids] if isinstance(cat_ids, int) else cat_ids

        if len(img_ids) == len(cat_ids) == len(area_range) == 0:
            anns = self.dataset["annotations"]
        else:
            if not len(img_ids) == 0:
                lists = [self.img_to_anns[img_id] for img_id in img_ids if img_id in self.img_to_anns]
                anns = list(itertools.chain.from_iterable(lists))
            else:
                anns = self.dataset["annotations"]
            anns = anns if len(cat_ids) == 0 else [ann for ann in anns if ann["category_id"] in cat_ids]
            anns = (
                anns if len(area_range) == 0 else [ann for ann in anns if area_range[0] < ann["area"] < area_range[1]]
            )
        if is_crowd is not None:
            ids = [ann["id"] for ann in anns if ann["iscrowd"] == is_crowd]
        else:
            ids = [ann["id"] for ann in anns]
        return ids

    def get_cat_ids(
        self,
        category_names: Optional[Union[str, Sequence[str]]] = None,
        super_category_names: Optional[Union[str, Sequence[str]]] = None,
        category_ids: Optional[Union[int, Sequence[int]]] = None,
    ) -> Sequence[int]:
        """
        Filtering parameters. default skips that filter.

        :param category_names: get cats for given cat names
        :param super_category_names: get cats for given super category names
        :param category_ids: get cats for given cat ids

        :return: ids: integer array of cat ids
        """

        if category_names is None:
            category_names = []
        if super_category_names is None:
            super_category_names = []
        if category_ids is None:
            category_ids = []

        category_names = [category_names] if isinstance(category_names, str) else category_names
        super_category_names = [super_category_names] if isinstance(super_category_names, str) else super_category_names
        category_ids = [category_ids] if isinstance(category_ids, int) else category_ids

        if len(category_names) == len(super_category_names) == len(category_ids) == 0:
            cats = self.dataset["categories"]
        else:
            cats = self.dataset["categories"]
            cats = cats if len(category_names) == 0 else [cat for cat in cats if cat["name"] in category_names]
            cats = (
                cats
                if len(super_category_names) == 0
                else [cat for cat in cats if cat["supercategory"] in super_category_names]
            )
            cats = cats if len(category_ids) == 0 else [cat for cat in cats if cat["id"] in category_ids]
        ids = [cat["id"] for cat in cats]
        return ids

    def get_image_ids(
        self, img_ids: Optional[Union[int, Sequence[int]]] = None, cat_ids: Optional[Union[int, Sequence[int]]] = None
    ) -> Sequence[int]:
        """
        Get img ids that satisfy given filter conditions.

        :param img_ids: get imgs for given ids
        :param cat_ids: get imgs with all given cats

        :return: ids: integer array of img ids
        """

        if img_ids is None:
            img_ids = []
        if cat_ids is None:
            cat_ids = []

        img_ids = [img_ids] if isinstance(img_ids, int) else img_ids
        cat_ids = [cat_ids] if isinstance(cat_ids, int) else cat_ids

        if len(img_ids) == len(cat_ids) == 0:
            ids = set(self.imgs.keys())
        else:
            ids = set(img_ids)
            for i, cat_id in enumerate(cat_ids):
                if i == 0 and len(ids) == 0:
                    ids = set(self.cat_to_imgs[cat_id])
                else:
                    ids &= set(self.cat_to_imgs[cat_id])
        return list(ids)

    def load_anns(self, ids: Optional[Union[int, Sequence[int]]] = None) -> List[JsonDict]:
        """
        Load anns with the specified ids.

        :param ids: integer ids specifying anns

        :return: anns: loaded ann objects
        """
        if ids is None:
            ids = []
        ids = [ids] if isinstance(ids, int) else ids

        return [self.anns[id] for id in ids]

    def load_cats(self, ids: Optional[Union[int, Sequence[int]]] = None) -> List[JsonDict]:
        """
        Load cats with the specified ids.

        :param ids: integer ids specifying cats

        :return: cats: loaded cat objects
        """
        if ids is None:
            ids = []
        ids = [ids] if isinstance(ids, int) else ids

        return [self.cats[idx] for idx in ids]

    def load_imgs(self, ids: Optional[Union[int, Sequence[int]]] = None) -> List[JsonDict]:
        """
        Load anns with the specified ids.

        :param ids: integer ids specifying img

        :return: imgs: loaded img objects
        """
        if ids is None:
            ids = []
        ids = [ids] if isinstance(ids, int) else ids

        return [self.imgs[idx] for idx in ids]


class SerializerCoco:
    """
    Class for serializing annotation files in Coco format. Coco comes in JSON format which is a priori not
    serialized. This class implements only the very basic methods to generate a dataflow. It wraps the coco class
    from pycocotools and assembles annotations that belong to the image. Note, that the conversion into the core
    `Image` has to be done by yourself.
    """

    @staticmethod
    def load(path: Pathlike, max_datapoints: Optional[int] = None) -> DataFlow:
        """
        Loads a .json file and generates a dataflow.

        **Example:**

                {'images':[img1,img2,...], 'annotations':[ann1,ann2,...],...}

            it will generate a dataflow with datapoints


                {'image':{'id',...},'annotations':[{'id':…,'bbox':...}]}

            for each single image id.

        :param max_datapoints: Will stop the iteration once max_datapoints have been streamed.
        :param path: a path to a .json file.
        :return: dataflow to iterate from
        """
        if not os.path.isfile(path):
            raise FileNotFoundError(path)
        file = os.path.split(path)[1]
        if not is_file_extension(file, ".json"):
            raise FileExtensionError(f"Expected .json file got {path}")

        with timed_operation("Start loading .json file and serializing"):
            coco = CocoParser(path)
            img_ids = coco.get_image_ids()
            imgs = coco.load_imgs(img_ids)

            with get_tqdm(total=len(imgs)) as status_bar:
                for img in imgs:
                    img["annotations"] = coco.img_to_anns[img["id"]]
                    status_bar.update()  # type: ignore

        df = CustomDataFromList(imgs, max_datapoints=max_datapoints)
        return df

    @staticmethod
    def save() -> None:
        """
        Not implemented
        """
        raise NotImplementedError


class SerializerPdfDoc:
    """
    Serialize a pdf document with an arbitrary number of pages.

    **Example:**

            df = SerializerPdfDoc.load("path/to/document.pdf")

        will yield datapoints:

            {"path": "path/to/document.pdf", "file_name" document_page_1.pdf, "pdf_bytes": b"some-bytes"}
    """

    @staticmethod
    def load(path: Pathlike, max_datapoints: Optional[int] = None) -> DataFlow:
        """
        Loads the document page wise and returns a dataflow accordingly.

        :param path: Path to the pdf document.
        :param max_datapoints: The maximum number of pages to stream.
        :return: A dict with structure {"path":... ,"file_name": ..., "pdf_bytes": ...}. The file name is a
                 concatenation of the physical file name and the current page number.
        """

        file_name = os.path.split(path)[1]
        prefix, suffix = os.path.splitext(file_name)
        df: DataFlow
        df = CustomDataFromIterable(PDFStreamer(path=path), max_datapoints=max_datapoints)
        df = MapData(
            df,
            lambda dp: {
                "path": path,
                "file_name": prefix + f"_{dp[1]}" + suffix,
                "pdf_bytes": dp[0],
                "page_number": dp[1],
                "document_id": get_uuid_from_str(prefix),
            },
        )
        return df

    @staticmethod
    def save(path: Pathlike) -> None:
        """
        Not implemented
        """
        raise NotImplementedError

    @staticmethod
    def split(path: Pathlike, path_target: Optional[Pathlike] = None, max_datapoint: Optional[int] = None) -> None:
        """
        Split a document into single pages.
        """
        if path_target is None:
            path_target, _ = os.path.split(path)
        if not os.path.isdir(path_target):
            raise NotADirectoryError(path)
        df = SerializerPdfDoc.load(path, max_datapoint)
        for dp in df:
            with open(os.path.join(path_target, dp["file_name"]), "wb") as page:
                page.write(dp["pdf_bytes"])



# -*- coding: utf-8 -*-
# File: test_fintabnet.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module datasets.instances.fintabnet
"""

from unittest.mock import MagicMock, patch

from pytest import mark

from deepdoctection.datasets import Fintabnet

from ...test_utils import collect_datapoint_from_dataflow, get_test_path
from .conftest import get_white_image


@mark.basic
@patch("deepdoctection.mapper.pubstruct.convert_pdf_bytes_to_np_array_v2", MagicMock(side_effect=get_white_image))
@patch("deepdoctection.mapper.pubstruct.load_bytes_from_pdf_file", MagicMock(return_value=b"\x01\x02"))
@patch("deepdoctection.datasets.instances.fintabnet.set_mp_spawn", MagicMock())
def test_dataset_fintabnet_returns_image() -> None:
    """
    test dataset fintabnet returns image
    """

    # Arrange
    fintabnet = Fintabnet()
    fintabnet.dataflow.get_workdir = get_test_path  # type: ignore
    fintabnet.dataflow.annotation_files = {"val": "test_file_fintab.jsonl"}
    df = fintabnet.dataflow.build(use_multi_proc=False)

    # Act
    df_list = collect_datapoint_from_dataflow(df)
    assert len(df_list) == 4


@mark.basic
@patch("deepdoctection.mapper.pubstruct.convert_pdf_bytes_to_np_array_v2", MagicMock(side_effect=get_white_image))
@patch("deepdoctection.mapper.pubstruct.load_bytes_from_pdf_file", MagicMock(return_value=b"\x01\x02"))
@patch("deepdoctection.datasets.instances.fintabnet.set_mp_spawn", MagicMock())
def test_dataset_fintabnet_with_load_image_returns_image() -> None:
    """
    test dataset fintabnet returns image
    """

    # Arrange
    fintabnet = Fintabnet()
    fintabnet.dataflow.get_workdir = get_test_path  # type: ignore
    fintabnet.dataflow.annotation_files = {"val": "test_file_fintab.jsonl"}
    df = fintabnet.dataflow.build(load_image=True, use_multi_proc=False)

    # Act
    df_list = collect_datapoint_from_dataflow(df)
    assert len(df_list) == 4
    assert df_list[0].image is not None



# -*- coding: utf-8 -*-
# File: tpstruct.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for mapping annotation for training environments
"""
import os.path
from typing import Optional, Sequence, Union

import numpy as np

from ..datapoint.annotation import ImageAnnotation
from ..datapoint.image import Image
from ..utils.detection_types import JsonDict
from ..utils.file_utils import tf_available
from ..utils.settings import ObjectTypes
from .maputils import curry

if tf_available():
    from tensorflow import convert_to_tensor, uint8  # type: ignore # pylint: disable=E0401
    from tensorflow.image import non_max_suppression  # type: ignore # pylint: disable=E0401


@curry
def image_to_tp_frcnn_training(
    dp: Image,
    add_mask: bool = False,
    category_names: Optional[Union[str, ObjectTypes, Sequence[Union[str, ObjectTypes]]]] = None,
) -> Optional[JsonDict]:
    """
    Maps an image to a dict to be consumed by Tensorpack Faster-RCNN bounding box detection. Note, that the returned
    dict will not suffice for training as gt for RPN and anchors still need to be created.

    :param dp: Image
    :param add_mask: True is not implemented (yet).
    :param category_names: A list of category names for training a model. Pass nothing to train with all annotations
    :return: Dict with `image`, `gt_boxes`, `gt_labels` and `file_name`, provided there are some detected objects in the
             image
    """

    output: JsonDict = {}
    if dp.image is not None:
        output["image"] = dp.image.astype("float32")
    anns = dp.get_annotation(category_names=category_names)
    all_boxes = []
    all_categories = []
    if not anns:
        return None

    for ann in anns:
        if ann.image is not None:
            box = ann.image.get_embedding(dp.image_id)
        else:
            box = ann.bounding_box
        if box is None:
            raise ValueError("BoundingBox cannot be None")
        all_boxes.append(box.to_list(mode="xyxy"))
        all_categories.append(ann.category_id)

        if add_mask:
            raise NotImplementedError

    output["gt_boxes"] = np.asarray(all_boxes, dtype="float32")
    output["gt_labels"] = np.asarray(all_categories, dtype="int32")
    if not os.path.isfile(dp.location) and dp.image is None:
        return None

    output["file_name"] = dp.location  # full path

    return output


def tf_nms_image_annotations(
    anns: Sequence[ImageAnnotation], threshold: float, image_id: Optional[str] = None
) -> Sequence[str]:
    """
    Processing given image annotations through NMS. This is useful, if you want to supress some specific image
    annotation, e.g. given by name or returned through different predictors. This is the tf version, for pt check
    `mapper.d2struct`

    :param anns: A sequence of ImageAnnotations. All annotations will be treated as if they belong to one category
    :param threshold: NMS threshold
    :param image_id: id in order to get the embedding bounding box

    :return: A list of annotation_ids that belong to the given input sequence and that survive the NMS process
    """
    if len(anns) == 1:
        return [anns[0].annotation_id]
    if not anns:
        return []
    ann_ids = np.array([ann.annotation_id for ann in anns], dtype="object")
    if image_id:
        boxes = convert_to_tensor(
            [ann.image.get_embedding(image_id).to_list(mode="xyxy") for ann in anns if ann.image is not None]
        )
        if not boxes.shape[0]:
            boxes = convert_to_tensor(
                [ann.bounding_box.to_list(mode="xyxy") for ann in anns if ann.bounding_box is not None]
            )
    else:
        boxes = convert_to_tensor(
            [ann.bounding_box.to_list(mode="xyxy") for ann in anns if ann.bounding_box is not None]
        )
    scores = convert_to_tensor([ann.score for ann in anns])
    class_mask = convert_to_tensor(len(boxes), dtype=uint8)
    keep = non_max_suppression(boxes, scores, class_mask, iou_threshold=threshold)
    ann_ids_keep = ann_ids[keep]
    if not isinstance(ann_ids_keep, str):
        return ann_ids_keep.tolist()
    return []



# -*- coding: utf-8 -*-
# File: test_d2_frcnn_train.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module train.d2_frcnn_train
"""

from pathlib import Path
from typing import Any
from unittest.mock import MagicMock, patch

from pytest import mark

from deepdoctection.datasets import DatasetBase
from deepdoctection.utils import detectron2_available

if detectron2_available():
    from deepdoctection.train.d2_frcnn_train import train_d2_faster_rcnn


def set_num_gpu_to_one() -> int:
    """
    set gpu number to one
    """
    return 1


@mark.requires_gpu
@patch("deepdoctection.train.d2_frcnn_train.get_num_gpu", MagicMock(side_effect=set_num_gpu_to_one))
@patch("deepdoctection.train.d2_frcnn_train.D2Trainer.train")
def test_train_faster_rcnn(
    patch_train: Any, path_to_d2_frcnn_yaml: str, test_dataset: DatasetBase, tmp_path: Path
) -> None:
    """
    test train faster rcnn runs until "trainer.train()"
    """

    # Arrange
    train_d2_faster_rcnn(
        path_config_yaml=path_to_d2_frcnn_yaml,
        dataset_train=test_dataset,
        path_weights="",
        log_dir=str(tmp_path),
        dataset_val=test_dataset,
        metric_name="coco",
        pipeline_component_name="ImageLayoutService",
    )

    # Assert
    patch_train.assert_called_once()



# -*- coding: utf-8 -*-
# File: test_d2detect.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module extern.d2detect
"""

from typing import Dict, List

from unittest.mock import MagicMock, patch
from pytest import mark, raises

from deepdoctection.utils.file_utils import pytorch_available, detectron2_available

from deepdoctection.extern.d2detect import D2FrcnnDetector
from deepdoctection.utils.detection_types import ImageType
from deepdoctection.utils.settings import ObjectTypes

if pytorch_available():
    import torch

if detectron2_available():
    from detectron2.structures import Instances, Boxes


def get_mock_instances() -> List[List[Dict[str,Instances]]]:
    """
    return Instances instance
    """
    pred_boxes = Boxes(torch.Tensor([[1.0, 1.6, 2.0, 4.6],[12.0, 12.0, 12.0, 12.0]]))
    scores = torch.Tensor([0.93,0.54])
    pred_classes = torch.Tensor([0,1]).to(torch.uint8)

    instance =  Instances((400,600))
    instance.pred_boxes = pred_boxes
    instance.scores = scores
    instance.pred_classes = pred_classes

    return [[{"instances": instance}]]


class TestD2FrcnnDetector:
    """
    Test D2FrcnnDetector
    """

    @staticmethod
    @mark.requires_pt
    @patch("deepdoctection.utils.file_utils.detectron2_available",MagicMock(return_value=False))
    def test_d2_does_not_build_when_d2_not_available(path_to_d2_frcnn_yaml: str,
                                                     categories: Dict[str,ObjectTypes]) -> None:
        """
        D2 FRCNN does only build when detectron2 is properly installed
        """

        # Arrange, Act & Assert
        with raises(ImportError):
            D2FrcnnDetector(path_yaml=path_to_d2_frcnn_yaml,path_weights="",categories=categories)

    @staticmethod
    @mark.requires_pt
    @patch("deepdoctection.extern.d2detect.D2FrcnnDetector.set_model", MagicMock(return_value=MagicMock))
    @patch("deepdoctection.extern.d2detect.D2FrcnnDetector._instantiate_d2_predictor", MagicMock())
    def test_d2_frcnn_predicts_image(path_to_d2_frcnn_yaml: str, categories: Dict[str,ObjectTypes],
                                     np_image: ImageType)-> None:
        """
        D2 FRCNN calls predict_image and post processes DetectionResult correctly, e.g. adding class names
        """

        # Arrange
        frcnn = D2FrcnnDetector(path_yaml=path_to_d2_frcnn_yaml,path_weights="",categories=categories)
        frcnn.d2_predictor = MagicMock(side_effect=get_mock_instances())
        # Act
        results = frcnn.predict(np_image)

        # Assert
        assert len(results)==2
        assert results[0].class_id == 1
        assert results[1].class_id == 2



# -*- coding: utf-8 -*-
# File: save.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for saving
"""

import json
from pathlib import Path
from typing import Optional

from cv2 import imwrite

from ..dataflow import DataFlow, MapData, SerializerJsonlines
from ..datapoint.convert import convert_b64_to_np_array
from ..datapoint.image import Image
from ..utils.detection_types import JsonDict, Pathlike
from ..utils.fs import mkdir_p


def dataflow_to_json(
    df: DataFlow,
    path: Pathlike,
    single_files: bool = False,
    file_name: Optional[str] = None,
    max_datapoints: Optional[int] = None,
    save_image_in_json: bool = True,
    highest_hierarchy_only: bool = False,
) -> None:
    """
    Save a dataflow consisting of `datapoint.Image` to a jsonl file. Each image will be dumped into a separate
    JSON object.

    :param df: Input dataflow
    :param path: Path to save the file(s) to
    :param single_files: will save image results to a single JSON file. If False all images of the dataflow will be d
                         dumped into a single jsonl file.
    :param file_name: file name, only needed for jsonl files
    :param max_datapoints: Will stop saving after dumping max_datapoint images.
    :param save_image_in_json: Will save the image to the JSON object
    :param highest_hierarchy_only: If True it will remove all image attributes of ImageAnnotations
    """
    if isinstance(path, str):
        path = Path(path)
    if single_files:
        mkdir_p(path)
    if not save_image_in_json:
        mkdir_p(path / "image")
    if highest_hierarchy_only:

        def _remove_hh(dp: Image) -> Image:
            dp.remove_image_from_lower_hierachy()
            return dp

        df = MapData(df, _remove_hh)
    df = MapData(df, lambda dp: dp.as_dict())

    def _path_to_str(dp: JsonDict) -> JsonDict:
        dp["location"] = str(dp["location"])
        return dp

    df = MapData(df, _path_to_str)
    df.reset_state()
    if single_files:
        for idx, dp in enumerate(df):
            if idx == max_datapoints:
                break
            target_file = path / (dp["file_name"].split(".")[0] + ".json")
            if not save_image_in_json:
                target_file_png = path / "image" / (dp["file_name"].split(".")[0] + ".png")
                image = dp.pop("_image")
                image = convert_b64_to_np_array(image)
                imwrite(str(target_file_png), image)

            with open(target_file, "w", encoding="UTF-8") as file:
                json.dump(dp, file)

    else:
        if not file_name:
            raise ValueError("If single_files is set to False must pass a valid file name for .jsonl file")
        SerializerJsonlines.save(df, path, file_name, max_datapoints)



# -*- coding: utf-8 -*-
# File: detection_types.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Typing sheet for the whole package
"""

import queue
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Protocol, Tuple, Type, TypeVar, Union

import numpy.typing as npt
import tqdm
from numpy import uint8


# Type for a general dataclass
class IsDataclass(Protocol):  # pylint: disable=R0903
    """
    type hint for general dataclass
    """

    __dataclass_fields__: Dict[Any, Any]


# Type for category dict of the DatasetCategories class
KeyValue = Union[str, int]

# Numpy image type
ImageType = npt.NDArray[uint8]

# typing for curry decorator
DP = TypeVar("DP")
S = TypeVar("S")
T = TypeVar("T")

# Some type hints that must be distinguished when runnning mypy and linters
if TYPE_CHECKING:
    QueueType = queue.Queue[Any]  # pylint: disable=E1136
    TqdmType = tqdm.tqdm[Any]  # pylint: disable=E1136
    BaseExceptionType = Type[BaseException]
else:
    BaseExceptionType = bool
    QueueType = queue.Queue
    TqdmType = tqdm.tqdm

JsonDict = Dict[str, Any]

# Type for requirements. A requirement is a Tuple of string and a callable that returns True if the requirement is
# available
Requirement = Tuple[str, bool, str]

# Pathlike, use this typing for everything where a path (absolute/relative) is involved
Pathlike = Union[str, Path]



# -*- coding: utf-8 -*-
# File: adapter.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for wrapping datasets into a pytorch dataset framework.
"""


from typing import Any, Callable, Iterator, Mapping, Optional, Union

from ..dataflow import CacheData, CustomDataFromList, MapData, RepeatedData
from ..datapoint.image import Image
from ..datasets.base import DatasetBase
from ..mapper.maputils import LabelSummarizer
from ..utils.detection_types import DP, JsonDict
from ..utils.file_utils import pytorch_available
from ..utils.logger import log_once, logger
from ..utils.settings import DatasetType, LayoutType, ObjectTypes, PageType, WordType
from ..utils.tqdm import get_tqdm
from .registry import get_dataset

if pytorch_available():
    from torch.utils.data import IterableDataset


class DatasetAdapter(IterableDataset):  # type: ignore
    """
    A helper class derived from `torch.utils.data.IterableDataset` to process datasets within
    pytorch frameworks (e.g. Detectron2). It wraps the dataset and defines the compulsory
    `__iter__` using  `dataflow.build` .

    DatasetAdapter is meant for training and will therefore produce an infinite number of datapoints
    by shuffling and restart iteration once the previous dataflow is exhausted.

    """

    def __init__(
        self,
        name_or_dataset: Union[str, DatasetBase],
        cache_dataset: bool,
        image_to_framework_func: Optional[Callable[[DP], Optional[JsonDict]]] = None,
        use_token_tag: bool = True,
        **build_kwargs: str,
    ) -> None:
        """
        :param name_or_dataset: Registered name of the dataset or an instance.
        :param cache_dataset: If set to true, it will cache the dataset (without loading images). If possible,
                              some statistics, e.g. number of specific labels will be printed.
        :param image_to_framework_func: A mapping function that converts image datapoints into the framework format
        :param use_token_tag: Will only be used for dataset_type="token_classification". If use_token_tag=True, will use
                              labels from sub category `WordType.token_tag` (with `B,I,O` suffix), otherwise
                              `WordType.token_class`.
        :param build_kwargs: optional parameters for defining the dataflow.
        """
        if isinstance(name_or_dataset, str):
            self.dataset = get_dataset(name_or_dataset)
        else:
            self.dataset = name_or_dataset

        df = self.dataset.dataflow.build(**build_kwargs)

        if cache_dataset:
            logger.info("Yielding dataflow into memory and create torch dataset")
            categories: Mapping[str, ObjectTypes] = {}
            _data_statistics = True
            if self.dataset.dataset_info.type in (DatasetType.object_detection, DatasetType.sequence_classification):
                categories = self.dataset.dataflow.categories.get_categories(filtered=True)
            elif self.dataset.dataset_info.type in (DatasetType.token_classification,):
                if use_token_tag:
                    categories = self.dataset.dataflow.categories.get_sub_categories(
                        categories=LayoutType.word,
                        sub_categories={LayoutType.word: [WordType.token_tag]},
                        keys=False,
                        values_as_dict=True,
                    )[LayoutType.word][WordType.token_tag]
                else:
                    categories = self.dataset.dataflow.categories.get_sub_categories(
                        categories=LayoutType.word, sub_categories={LayoutType.word: [WordType.token_class]}, keys=False
                    )[LayoutType.word][WordType.token_class]
            else:
                logger.info(
                    "dataset is of type %s. Cannot generate statistics for this type of dataset",
                    self.dataset.dataset_info.type,
                )
                _data_statistics = False

            if _data_statistics:
                summarizer = LabelSummarizer(categories)

            df.reset_state()
            max_datapoints_str = build_kwargs.get("max_datapoints")

            if max_datapoints_str is not None:
                max_datapoints = int(max_datapoints_str)
            else:
                max_datapoints = None

            datapoints = []
            with get_tqdm(total=max_datapoints) as status_bar:
                for dp in df:
                    if dp.image is not None:
                        log_once(
                            "Datapoint have images as np arrays stored and they will be loaded into memory. "
                            "To avoid OOM set 'load_image'=False in dataflow build config. This will load "
                            "images when needed and reduce memory costs!!!",
                            "warn",
                        )
                    if self.dataset.dataset_info.type == DatasetType.object_detection:
                        anns = dp.get_annotation()
                        cat_ids = [int(ann.category_id) for ann in anns]

                    elif self.dataset.dataset_info.type == DatasetType.sequence_classification:
                        cat_ids = dp.summary.get_sub_category(PageType.document_type).category_id

                    elif self.dataset.dataset_info.type == DatasetType.token_classification:
                        anns = dp.get_annotation()
                        if use_token_tag:
                            cat_ids = [ann.get_sub_category(WordType.token_tag).category_id for ann in anns]
                        else:
                            cat_ids = [ann.get_sub_category(WordType.token_class).category_id for ann in anns]

                    if _data_statistics:
                        summarizer.dump(cat_ids)

                    datapoints.append(dp)
                    status_bar.update()  # type: ignore

            if _data_statistics:
                summarizer.print_summary_histogram()
            self.number_datapoints = len(datapoints)

            df = CustomDataFromList(datapoints, shuffle=True)
            if not image_to_framework_func:
                df = RepeatedData(df, -1)
            else:
                df_list = CacheData(df).get_cache()
                df = CustomDataFromList(df_list, shuffle=True)
                df = RepeatedData(df, -1)

        if image_to_framework_func:
            df = MapData(df, image_to_framework_func)

        self.df = df
        self.df.reset_state()

    def __iter__(self) -> Iterator[Image]:
        return iter(self.df)

    def __len__(self) -> int:
        if self.number_datapoints:
            return self.number_datapoints
        return len(self.df)

    def __getitem__(self, item: Any) -> None:
        raise NotImplementedError



# -*- coding: utf-8 -*-
# File: pdf_utils.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module with pdf processing tools
"""
import os
import platform
import subprocess
import sys
from errno import ENOENT
from io import BytesIO
from shutil import copyfile
from typing import Generator, List, Optional, Tuple

from cv2 import IMREAD_COLOR, imread
from numpy import uint8
from PyPDF2 import PdfFileReader, PdfFileWriter, errors

from .context import save_tmp_file, timeout_manager
from .detection_types import ImageType, Pathlike
from .file_utils import PopplerNotFound, pdf_to_cairo_available, pdf_to_ppm_available, qpdf_available
from .logger import logger
from .utils import FileExtensionError, is_file_extension

__all__ = ["decrypt_pdf_document", "get_pdf_file_reader", "get_pdf_file_writer", "PDFStreamer", "pdf_to_np_array"]


def decrypt_pdf_document(path: Pathlike) -> bool:
    """
    Decrypting a pdf. As copying a pdf document removes the password that protects pdf, this method
    generates a copy and decrypts the copy using qpdf. The result is saved as the original
    document.

    qpdf: <http://qpdf.sourceforge.net/>

    Note, that this is decryption does not work, if the pdf has a readable protection, in which case we do not
    provide any solution.

    :param path: A path to the pdf file
    :return: True if document has been successfully decrypted
    """

    if qpdf_available():
        path_base, file_name = os.path.split(path)
        file_name_tmp = os.path.splitext(file_name)[0] + "tmp.pdf"
        path_tmp = os.path.join(path_base, file_name_tmp)
        copyfile(path, path_tmp)
        cmd_str = f"qpdf --password='' --decrypt {path_tmp} {os.path.join(path_base, path)}"
        response = os.system(cmd_str)
        os.remove(path_tmp)
        if not response:
            return True
    else:
        logger.info("qpdf is not installed. If the document must be decrypted please ensure that it is installed")
    return False


def get_pdf_file_reader(path: Pathlike) -> PdfFileReader:
    """
    Creates a file reader object from a pdf document. Will try to decrypt the document if it is
    encrypted. (See `decrypt_pdf_document` to understand what is meant with "decrypt").

    :param path: A path to a pdf document
    :return: A file reader object from which you can iterate through the document.
    """

    if not os.path.isfile(path):
        raise FileNotFoundError(str(path))
    file_name = os.path.split(path)[1]
    if not is_file_extension(file_name, ".pdf"):
        raise FileExtensionError(f"must be a pdf file: {file_name}")

    with open(path, "rb") as file:
        qpdf_called = False
        try:
            input_pdf_as_bytes = PdfFileReader(file)
        except (errors.PdfReadError, AttributeError):
            _ = decrypt_pdf_document(path)
            qpdf_called = True

        if not qpdf_called:
            if input_pdf_as_bytes.isEncrypted:
                is_decrypted = decrypt_pdf_document(path)
                if not is_decrypted:
                    logger.error("pdf document %s cannot be decrypted and therefore cannot be processed further.", path)
                    sys.exit()

    file_reader = PdfFileReader(open(path, "rb"))  # pylint: disable=R1732
    return file_reader


def get_pdf_file_writer() -> PdfFileWriter:
    """
    `PdfFileWriter` instance
    """
    return PdfFileWriter()


class PDFStreamer:
    """
    A class for streaming pdf documents as bytes objects. Build as a generator, it is possible to load the document
    iteratively into memory. Uses py2pdf FileReader and FileWriter.

    **Example:**

             df = dataflow.DataFromIterable.PDFStreamer(path=path)
             df.reset_state()

             for page in df:
                ... # do whatever you like

    """

    def __init__(self, path: Pathlike) -> None:
        """
        :param path: to a pdf.
        """
        self.file_reader = get_pdf_file_reader(path)
        self.file_writer = PdfFileWriter()

    def __len__(self) -> int:
        return self.file_reader.getNumPages()

    def __iter__(self) -> Generator[Tuple[bytes, int], None, None]:
        for k in range(len(self)):
            buffer = BytesIO()
            writer = PdfFileWriter()
            writer.addPage(self.file_reader.getPage(k))
            writer.write(buffer)
            yield buffer.getvalue(), k


# The following functions are modified versions from the Python poppler wrapper
# https://github.com/Belval/pdf2image/blob/master/pdf2image/pdf2image.py


def _input_to_cli_str(
    input_file_name: Pathlike, output_file_name: Pathlike, dpi: int, size: Optional[Tuple[int, int]] = None
) -> List[str]:

    cmd_args: List[str] = []

    if pdf_to_ppm_available():
        command = "pdftoppm"
    elif pdf_to_cairo_available():
        command = "pdftocairo"
    else:
        raise PopplerNotFound("Poppler not found. Please install or add to your PATH.")

    if platform.system() == "Windows":
        command = command + ".exe"
    cmd_args.append(command)
    cmd_args.extend(["-r", str(dpi), str(input_file_name)])
    cmd_args.append("-png")
    cmd_args.append(str(output_file_name))

    if size:
        assert len(size) == 2, size
        assert isinstance(size[0], int) and isinstance(size[1], int), size
        cmd_args.extend(["-scale-to-x", str(size[0])])
        cmd_args.extend(["-scale-to-y", str(size[1])])

    return cmd_args


class PopplerError(RuntimeError):
    """
    Poppler Error
    """

    def __init__(self, status: int, message: str) -> None:
        super().__init__()
        self.status = status
        self.message = message
        self.args = (status, message)


def _run_poppler(poppler_args: List[str]) -> None:
    try:
        proc = subprocess.Popen(poppler_args)  # pylint: disable=R1732
    except OSError as error:
        if error.errno != ENOENT:
            raise error from error
        raise PopplerNotFound("Poppler not found. Please install or add to your PATH.") from error

    with timeout_manager(proc, 0):
        if proc.returncode:
            raise PopplerError(status=proc.returncode, message="Syntax Error: PDF cannot be read with Poppler")


def pdf_to_np_array(pdf_bytes: bytes, size: Optional[Tuple[int, int]] = None, dpi: int = 200) -> ImageType:
    """
    Convert a single pdf page from its byte representation to a numpy array. This function will save the pdf as to a tmp
    file and then call poppler via `pdftoppm` resp. `pdftocairo` if the former is not available.

    :param pdf_bytes: Bytes representing the PDF file
    :param size: Size of the resulting image(s), uses (width, height) standard
    :param dpi:  Image quality in DPI/dots-per-inch (default 200)
    :return: numpy array
    """

    with save_tmp_file(pdf_bytes, "pdf_") as (tmp_name, input_file_name):
        _run_poppler(_input_to_cli_str(input_file_name, tmp_name, dpi, size))
        image = imread(tmp_name + "-1.png", IMREAD_COLOR)

    return image.astype(uint8)



# -*- coding: utf-8 -*-
# File: test_registry.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module datasets.instances.registry
"""

from unittest.mock import MagicMock

from pytest import mark

from deepdoctection.datasets.base import DatasetBase
from deepdoctection.datasets.dataflow_builder import DataFlowBaseBuilder
from deepdoctection.datasets.info import DatasetCategories, DatasetInfo
from deepdoctection.datasets.registry import dataset_registry, get_dataset


@mark.basic
def test_dataset_registry_has_all_build_in_datasets_registered() -> None:
    """
    test dataset registry has all pipeline components registered
    """
    assert len(dataset_registry.get_all()) == 12


@mark.basic
def test_dataset_registry_registered_new_dataset() -> None:
    """
    test, that the new generated dataset component "TestDataset" can be registered and retrieved from registry
    """

    @dataset_registry.register("testdata")
    class TestDataset(DatasetBase):
        """
        TestDataset
        """

        @classmethod
        def _info(cls) -> DatasetInfo:
            """
            Processing an image through the whole pipeline component.
            """
            return MagicMock()

        def _categories(self) -> DatasetCategories:
            """
            _categories
            """
            return MagicMock()

        def _builder(self) -> DataFlowBaseBuilder:
            """
            _builder
            """
            return MagicMock()

    # Act
    test = get_dataset("testdata")

    # Assert
    assert isinstance(test, TestDataset)



# -*- coding: utf-8 -*-
# File: texocr.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
AWS Textract OCR engine for text extraction
"""

from typing import List

from ..datapoint.convert import convert_np_array_to_b64_b
from ..utils.detection_types import ImageType, JsonDict, Requirement
from ..utils.file_utils import boto3_available, get_aws_requirement, get_boto3_requirement
from ..utils.settings import LayoutType, ObjectTypes
from .base import DetectionResult, ObjectDetector, PredictorBase

if boto3_available():
    import boto3  # type:ignore


def _textract_to_detectresult(response: JsonDict, width: int, height: int, text_lines: bool) -> List[DetectionResult]:
    all_results: List[DetectionResult] = []
    blocks = response.get("Blocks")

    if blocks:
        for block in blocks:
            if (block["BlockType"] in "WORD") or (block["BlockType"] in "LINE" and text_lines):
                word = DetectionResult(
                    box=[
                        block["Geometry"]["Polygon"][0]["X"] * width,
                        block["Geometry"]["Polygon"][0]["Y"] * height,
                        block["Geometry"]["Polygon"][2]["X"] * width,
                        block["Geometry"]["Polygon"][2]["Y"] * height,
                    ],
                    score=block["Confidence"] / 100,
                    text=block["Text"],
                    class_id=1 if block["BlockType"] == "WORD" else 2,
                    class_name=LayoutType.word if block["BlockType"] == "WORD" else LayoutType.line,
                )
                all_results.append(word)

    return all_results


def predict_text(np_img: ImageType, client, text_lines: bool) -> List[DetectionResult]:  # type: ignore
    """
    Calls AWS Textract client (`detect_document_text`) and returns plain OCR results.
    AWS account required.

    :param client: botocore textract client
    :param np_img: Image in np.array.
    :param text_lines: If True, it will return DetectionResults of Text lines as well.
    :return: A list of textract extractions wrapped in DetectionResult
    """

    width, height = np_img.shape[1], np_img.shape[0]
    b_img = convert_np_array_to_b64_b(np_img)
    response = client.detect_document_text(Document={"Bytes": b_img})
    all_results = _textract_to_detectresult(response, width, height, text_lines)
    return all_results


class TextractOcrDetector(ObjectDetector):
    """
    Text object detector based on AWS Textract OCR engine. Note that an AWS account as well as some additional
    installations are required, i.e AWS CLI and boto3. Note further, that the service is not free of charge. Additional
    information can be found at: <https://docs.aws.amazon.com/textract/?id=docs_gateway> .

    The detector only calls the base OCR engine and does not return additional Textract document analysis features.

        textract_predictor = TextractOcrDetector()
        detection_result = textract_predictor.predict(bgr_image_as_np_array)

    or

        textract_predictor = TextractOcrDetector()
        text_extract = TextExtractionService(textract_predictor)

        pipe = DoctectionPipe([text_extract])
        df = pipe.analyze(path="path/to/document.pdf")

        for dp in df:
            ...

    """

    def __init__(self, text_lines: bool = False) -> None:
        """
        :param text_lines: If True, it will return DetectionResults of Text lines as well.
        """
        self.name = "textract"
        self.text_lines = text_lines
        self.client = boto3.client("textract")
        if self.text_lines:
            self.categories = {"1": LayoutType.word, "2": LayoutType.line}
        else:
            self.categories = {"1": LayoutType.word}

    def predict(self, np_img: ImageType) -> List[DetectionResult]:
        """
        Transfer of a numpy array and call textract client. Return of the detection results.

        :param np_img: image as numpy array
        :return: A list of DetectionResult
        """

        return predict_text(np_img, self.client, self.text_lines)

    @classmethod
    def get_requirements(cls) -> List[Requirement]:
        return [get_aws_requirement(), get_boto3_requirement()]

    def clone(self) -> PredictorBase:
        return self.__class__()

    def possible_categories(self) -> List[ObjectTypes]:
        if self.text_lines:
            return [LayoutType.word, LayoutType.line]
        return [LayoutType.word]
